# flake8: noqa
from .jurisdiction import State, JurisdictionScraper
from .popolo import Organization
from .vote_event import VoteEvent
from .bill import Bill
from .event import Event, calculate_window
from .base import Scraper, BaseBillScraper
from ..exceptions import EmptyScrape



from ..models import State, Chamber, simple_numbered_districts

MO = State(
    name="Missouri",
    abbr="MO",
    capital="Jefferson City",
    capital_tz="America/Chicago",
    fips="29",
    unicameral=False,
    legislature_name="Missouri General Assembly",
    legislature_organization_id="ocd-organization/997c6150-6c45-4178-82dd-e111c0029554",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/dd35cadb-0d6f-4a80-a84a-6ab08c2701b8",
    division_id="ocd-division/country:us/state:mo",
    jurisdiction_id="ocd-jurisdiction/country:us/state:mo/government",
    url="http://www.moga.mo.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/b028938f-599d-4b3a-8d53-fdb9de12e759",
        num_seats=163,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:mo", "lower", 163
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/8ff25b39-cd33-4cee-a2a9-d88b80b2c6fe",
        num_seats=34,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:mo", "upper", 34
        ),
    ),
)



import os
import pytest  # type: ignore
from unittest.mock import patch
from pathlib import Path
from pydantic import ValidationError
from openstates.cli.committees import CommitteeDir, PersonMatcher
from openstates.models.committees import (
    Committee,
    Link,
    ScrapeCommittee,
    Membership,
)

JURISDICTION_ID = "ocd-jurisdiction/country:us/state:wa/government"

TEST_DATA_PATH = Path(__file__).parent / "testdata"


def test_validation_chamber():
    ScrapeCommittee(name="Education", chamber="lower")
    ScrapeCommittee(name="Education", chamber="upper")
    ScrapeCommittee(name="Education", chamber="legislature")
    with pytest.raises(ValidationError):
        ScrapeCommittee(name="Education", chamber="joint")


def test_parent_validation_subcommittee():
    # subcommittees can be any string...
    ScrapeCommittee(
        name="Pre-K", chamber="upper", parent="Education", classification="subcommittee"
    )
    # must set parent if subcommittee
    with pytest.raises(ValidationError):
        ScrapeCommittee(name="Pre-K", chamber="upper", classification="subcommittee")
    # must set can't set parent without subcommittees
    with pytest.raises(ValidationError):
        ScrapeCommittee(name="Pre-K", chamber="upper", parent="Education")


@pytest.fixture
def person_matcher():
    pm = PersonMatcher("wa", TEST_DATA_PATH / "no-such-dir")
    pm.add_name("lower", "Jones", "ocd-person/00000000-0000-0000-0000-111111111111")
    pm.add_name("lower", "Nguyen", "ocd-person/00000000-0000-0000-0000-222222222222")
    pm.add_name("lower", "Green", "ocd-person/00000000-0000-0000-0000-333333333333")
    # two Cristobals
    pm.add_name("lower", "Cristobal", "ocd-person/00000000-0000-0000-0000-000888888888")
    pm.add_name("lower", "Cristobal", "ocd-person/00000000-0000-0000-0000-000999999999")
    return pm


def test_person_matcher_match(person_matcher):
    assert (
        person_matcher.match("lower", "Jones")
        == "ocd-person/00000000-0000-0000-0000-111111111111"
    )
    assert (
        person_matcher.match("lower", "Nguyen")
        == "ocd-person/00000000-0000-0000-0000-222222222222"
    )
    # two matches
    assert person_matcher.match("lower", "Cristobal") is None
    # no matches
    assert person_matcher.match("lower", "Gordy") is None


def test_merge_committees_name():
    comdir = CommitteeDir(abbr="wa", directory=TEST_DATA_PATH / "committees")
    id_one = "ocd-organization/00000000-0000-0000-0000-000000000001"
    id_two = "ocd-organization/00000000-0000-0000-0000-000000000002"
    c1 = Committee(
        id=id_one, jurisdiction=JURISDICTION_ID, chamber="upper", name="Education"
    )
    c2 = Committee(
        id=id_two,
        jurisdiction=JURISDICTION_ID,
        chamber="upper",
        name="Education & Children",
    )
    merged = comdir.merge_committees(c1, c2)
    assert merged.id == c1.id
    assert merged.name == c2.name


def test_merge_committees_invalid():
    comdir = CommitteeDir(abbr="wa", directory=TEST_DATA_PATH / "committees")

    id_one = "ocd-organization/00000000-0000-0000-0000-000000000001"
    id_two = "ocd-organization/00000000-0000-0000-0000-000000000002"
    c1 = Committee(
        id=id_one, jurisdiction=JURISDICTION_ID, chamber="upper", name="Education"
    )
    c2 = Committee(
        id=id_two,
        jurisdiction=JURISDICTION_ID,
        chamber="lower",
        name="Education & Children",
    )
    with pytest.raises(ValueError):
        comdir.merge_committees(c1, c2)


def test_merge_committees_links():
    comdir = CommitteeDir(abbr="wa", directory=TEST_DATA_PATH / "committees")
    id_one = "ocd-organization/00000000-0000-0000-0000-000000000001"
    id_two = "ocd-organization/00000000-0000-0000-0000-000000000002"
    c1 = Committee(
        id=id_one,
        jurisdiction=JURISDICTION_ID,
        chamber="upper",
        name="Education",
        links=[
            Link(url="https://example.com/1"),
            Link(url="https://example.com/2"),
        ],
    )
    c2 = Committee(
        id=id_two,
        jurisdiction=JURISDICTION_ID,
        chamber="upper",
        name="Education & Children",
        links=[
            Link(url="https://example.com/1", note="first"),
            Link(url="https://example.com/3"),
        ],
    )
    merged = comdir.merge_committees(c1, c2)
    assert merged.links == [
        Link(url="https://example.com/1", note="first"),
        Link(url="https://example.com/2"),
        Link(url="https://example.com/3"),
    ]


def test_merge_committees_members():
    comdir = CommitteeDir(abbr="wa", directory=TEST_DATA_PATH / "committees")
    id_one = "ocd-organization/00000000-0000-0000-0000-000000000001"
    id_two = "ocd-organization/00000000-0000-0000-0000-000000000002"
    person_id = "ocd-person/00000000-0000-0000-0000-000000000002"
    c1 = Committee(
        id=id_one,
        jurisdiction=JURISDICTION_ID,
        chamber="upper",
        name="Education",
        members=[
            Membership(name="Amy", role="chair"),
            Membership(name="Bo", role="chair"),
        ],
    )
    c2 = Committee(
        id=id_two,
        jurisdiction=JURISDICTION_ID,
        chamber="upper",
        name="Education & Children",
        members=[
            Membership(name="Amy", role="chair", person_id=person_id),
            Membership(name="Charlize", role="member"),
        ],
    )
    merged = comdir.merge_committees(c1, c2)
    assert merged.members == [
        Membership(name="Amy", role="chair", person_id=person_id),
        Membership(name="Bo", role="chair"),
        Membership(name="Charlize", role="member"),
    ]


def test_load_data():
    comdir = CommitteeDir(abbr="wa", directory=TEST_DATA_PATH / "committees")

    assert len(comdir.coms_by_chamber_and_name["lower"]) == 3
    assert len(comdir.coms_by_chamber_and_name["upper"]) == 1
    assert comdir.errors == []


def test_load_data_with_errors():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "broken-committees",
        raise_errors=False,
    )

    assert len(comdir.coms_by_chamber_and_name["lower"]) == 0
    assert len(comdir.coms_by_chamber_and_name["upper"]) == 0
    assert len(comdir.errors) == 2
    # error order isn't deterministic
    path0, msg0 = comdir.errors[0]
    path1, msg1 = comdir.errors[1]
    if "lower" in str(path0):
        msg0, msg1 = msg1, msg0
    assert "8 validation errors" in str(msg0)
    assert "members -> 3 -> who" in str(msg0)
    assert "members -> 3 -> name" in str(msg0)
    assert "2 validation errors" in str(msg1)
    assert "not a valid enumeration member" in str(msg1)
    assert "extra fields not permitted" in str(msg1)


def test_load_data_with_errors_raised():
    # default is to raise error right away, test_load_data_with_errors catches errors for linter
    with pytest.raises(ValidationError):
        CommitteeDir(
            abbr="wa",
            directory=TEST_DATA_PATH / "broken-committees",
        )


def test_get_new_filename():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    simple = Committee(
        id="ocd-organization/00001111-2222-3333-4444-555566667777",
        jurisdiction=JURISDICTION_ID,
        name="Simple",
        chamber="lower",
    )
    longer = Committee(
        id="ocd-organization/00001111-2222-3333-4444-999999999999",
        jurisdiction=JURISDICTION_ID,
        name="Ways, Means & Taxes",
        chamber="upper",
    )
    assert (
        comdir.get_new_filename(simple)
        == "lower-Simple-00001111-2222-3333-4444-555566667777.yml"
    )
    assert (
        comdir.get_new_filename(longer)
        == "upper-Ways-Means--Taxes-00001111-2222-3333-4444-999999999999.yml"
    )


def test_get_filename_by_id():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    assert comdir.get_filename_by_id(
        "ocd-organization/11111111-2222-3333-4444-111111111111"
    ) == (
        TEST_DATA_PATH
        / "committees/lower-Agriculture-11111111-2222-3333-4444-111111111111.yml"
    )

    with pytest.raises(FileNotFoundError):
        comdir.get_filename_by_id(
            "ocd-organization/99999999-9999-9999-9999-999999999999"
        )


def test_get_filename_by_name():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    assert comdir.get_filename_by_name("lower", "Agriculture") == (
        TEST_DATA_PATH
        / "committees/lower-Agriculture-11111111-2222-3333-4444-111111111111.yml"
    )

    with pytest.raises(FileNotFoundError):
        comdir.get_filename_by_name("lower", "Weird")


# TODO: test_save_committee, test_add_committee


def test_add_committee():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    with patch.object(comdir, "save_committee") as patch_obj:
        sc = ScrapeCommittee(chamber="lower", name="New Business")
        comdir.add_committee(sc)
        full_com = comdir.coms_by_chamber_and_name[sc.chamber][sc.name]
        assert full_com.name == sc.name
        assert full_com.id.startswith("ocd-organization")
        assert full_com.jurisdiction == JURISDICTION_ID
        assert patch_obj.called_once_with(full_com)


def test_ingest_scraped_json():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    os.environ["OS_PEOPLE_DIRECTORY"] = str(TEST_DATA_PATH)
    committees = comdir.ingest_scraped_json(TEST_DATA_PATH / "scraped-committees")
    assert len(committees) == 2
    assert {"Judiciary 2", "Judiciary 4"} == {c.name for c in committees}


def test_ingest_scraped_json_names_resolved():
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    richardson_id = "ocd-person/11111111-0000-0000-0000-555555555555"
    comdir.person_matcher = PersonMatcher("wa", TEST_DATA_PATH / "no-such-dir")
    comdir.person_matcher.add_name("lower", "Richardson", richardson_id)
    committees = comdir.ingest_scraped_json(TEST_DATA_PATH / "scraped-committees")
    assert len(committees) == 2
    committees = sorted(committees, key=lambda c: c.name)
    assert committees[0].name == "Judiciary 2"
    # ensure that names are matched
    assert committees[0].members[0].name == "Richardson"
    assert committees[0].members[0].role == "chair"
    assert committees[0].members[0].person_id == richardson_id
    assert committees[1].name == "Judiciary 4"


def test_get_merge_plan_by_chamber(person_matcher):
    comdir = CommitteeDir(
        abbr="wa",
        directory=TEST_DATA_PATH / "committees",
    )
    comdir.person_matcher = person_matcher

    newdata = [
        # identical
        ScrapeCommittee(
            name="Education",
            chamber="lower",
            sources=[Link(url="https://example.com/committee")],
            members=[
                Membership(name="Jones", role="chair"),
                Membership(name="Nguyen", role="co-chair"),
                Membership(name="Green", role="member"),
                Membership(name="Cristobal", role="member"),
            ],
        ),
        # new
        ScrapeCommittee(
            name="Science",
            chamber="lower",
            sources=[Link(url="https://example.com/committee")],
            members=[
                Membership(name="Jones", role="chair"),
                Membership(name="Nguyen", role="co-chair"),
            ],
        ),
        # changed
        ScrapeCommittee(
            name="Rules",
            chamber="lower",
            sources=[Link(url="https://example.com/committee")],
            members=[
                Membership(name="Fox", role="chair"),
                Membership(name="Fawkes", role="co-chair"),
                Membership(name="Faux", role="member"),
            ],
        ),
    ]

    plan = comdir.get_merge_plan_by_chamber("lower", newdata)
    assert plan.names_to_add == {"Science"}
    assert plan.names_to_remove == {"Agriculture"}
    assert plan.same == 1  # Edcuation
    assert len(plan.to_merge) == 1
    old, new = plan.to_merge[0]
    assert old.name == new.name == "Rules"
    assert len(old.members) < len(new.members)



import re
import datetime
import typing
from pathlib import Path
import yaml
from pydantic import BaseModel as PydanticBaseModel, validator
from openstates.metadata import lookup

JURISDICTION_RE = re.compile(
    r"ocd-jurisdiction/country:us/(state|district|territory):\w\w/(place|county):[a-z_]+/government"
)
ORG_ID_RE = re.compile(
    r"^ocd-organization/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
)
PERSON_ID_RE = re.compile(
    r"^ocd-person/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
)
DATE_RE = re.compile(r"^\d{4}(-\d{2}(-\d{2})?)?$")


def validate_str_no_newline(v: typing.Any) -> str:
    if isinstance(v, str) and "\n" in v:
        raise ValueError("must be a string without newline")
    # normalize spaces
    return re.sub(r"\s+", " ", v).strip()


def validate_fuzzy_date(v: typing.Any) -> typing.Union[datetime.date, str]:
    if isinstance(v, datetime.date):
        return v
    elif isinstance(v, str) and (v == "" or DATE_RE.match(v)):
        return v
    else:
        raise ValueError("invalid date")


def validate_ocd_person(v: typing.Any) -> str:
    if isinstance(v, str) and not PERSON_ID_RE.match(v):
        raise ValueError("must match ocd-person/UUID format")
    return v


def validate_ocd_jurisdiction(v: typing.Any) -> str:
    try:
        lookup(jurisdiction_id=v)
    except KeyError:
        if not JURISDICTION_RE.match(v):
            raise ValueError(f"invalid jurisdiction_id {v}")
    return v


def validate_ocd_organization(v: str) -> str:
    if isinstance(v, str) and not ORG_ID_RE.match(v):
        raise ValueError("must match ocd-organization/UUID format")
    return v


def validate_url(v: str) -> str:
    if not v.startswith(("http://", "https://", "ftp://")):
        raise ValueError("URL must start with protocol")
    return v


BaseT = typing.TypeVar("BaseT", bound="BaseModel")


class BaseModel(PydanticBaseModel):
    class Config:
        anystr_strip_whitespace = True
        extra = "forbid"
        validate_assignment = True
        use_enum_values = True

    def to_dict(self) -> dict[str, typing.Any]:
        # TODO: replace this with first class pydantic support in spatula
        return self.dict()

    @classmethod
    def load_yaml(cls, filename: Path) -> BaseT:
        with open(filename) as file:
            data = yaml.safe_load(file)
            return cls(**data)  # type: ignore


class Link(BaseModel):
    url: str
    note: str = ""

    _validate_note = validator("note", allow_reuse=True)(validate_str_no_newline)
    _validate_url = validator("url", allow_reuse=True)(validate_url)


class TimeScoped(BaseModel):
    start_date: typing.Union[str, datetime.date] = ""
    end_date: typing.Union[str, datetime.date] = ""

    _validate_dates = validator("start_date", "end_date", allow_reuse=True)(
        validate_fuzzy_date
    )

    def is_active(self) -> bool:
        date = datetime.datetime.utcnow().date().isoformat()
        return (self.end_date == "" or str(self.end_date) > date) and (
            self.start_date == "" or str(self.start_date) <= date
        )


class OtherName(TimeScoped):
    name: str

    _validate_strs = validator("name", allow_reuse=True)(validate_str_no_newline)


class OtherIdentifier(TimeScoped):
    scheme: str
    identifier: str

    _validate_strs = validator("scheme", "identifier", allow_reuse=True)(
        validate_str_no_newline
    )



from ..models import State, Chamber, simple_numbered_districts

ND = State(
    name="North Dakota",
    abbr="ND",
    capital="Bismarck",
    capital_tz="America/North_Dakota/Center",
    fips="38",
    unicameral=False,
    legislature_name="North Dakota Legislative Assembly",
    legislature_organization_id="ocd-organization/aa301a79-0ad8-4905-818c-0cd5f1ce5946",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/5ba1c008-42e8-55ad-864e-e95b318d7ce4",
    division_id="ocd-division/country:us/state:nd",
    jurisdiction_id="ocd-jurisdiction/country:us/state:nd/government",
    url="http://www.legis.nd.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/b9f9b32c-a2cc-4cca-941f-862aecc30617",
        num_seats=94,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nd", "lower", 47, num_seats=2
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/79645ed7-6baa-4842-ac97-09b444c4020b",
        num_seats=47,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nd", "upper", 47
        ),
    ),
)



from django.contrib import admin
from .. import models
from .base import ModelAdmin, ReadOnlyTabularInline


@admin.register(models.Division)
class DivisionAdmin(ModelAdmin):
    list_display = ("name", "id")
    search_fields = list_display
    fields = readonly_fields = ("id", "name", "redirect", "country")
    ordering = ("id",)


class LegislativeSessionInline(ReadOnlyTabularInline):
    model = models.LegislativeSession
    readonly_fields = ("identifier", "name", "classification", "start_date", "end_date")
    ordering = ("-identifier",)


@admin.register(models.Jurisdiction)
class JurisdictionAdmin(ModelAdmin):
    list_display = ("name", "id")
    readonly_fields = fields = (
        "id",
        "name",
        "division",
        "classification",
        "extras",
        "url",
    )
    ordering = ("id",)
    inlines = [LegislativeSessionInline]



import sys
import csv
import typing
import datetime
import itertools
from collections import Counter, defaultdict
from pathlib import Path
import click
import boto3  # type: ignore
import yaml
from django.db import transaction  # type: ignore
from ..utils import abbr_to_jid
from ..utils.django import init_django  # type: ignore
from ..models.people import Person, Role, Party, Link
from ..utils.people import (
    ocd_uuid,
    get_data_path,
    dump_obj,
    get_all_abbreviations,
    download_state_images,
    load_municipalities,
)
from ..utils.people.retire import retire_person, add_vacancy, retire_file
from ..utils.people.lint_people import Validator, BadVacancy, PersonType
from ..utils.people.to_database import (
    create_municipalities,
    create_parties,
    load_person,
    CancelTransaction,
)
from ..utils.people.merge import process_scrape_dir, incoming_merge


OPTIONAL_FIELD_SET = {
    "sort_name",
    "given_name",
    "family_name",
    "gender",
    "summary",
    "biography",
    "birth_date",
    "image",
    "email",
    "other_names",
}


class Summarizer:
    def __init__(self) -> None:
        self.person_count = 0
        self.optional_fields: Counter[str] = Counter()
        self.extra_counts: Counter[str] = Counter()
        self.contact_counts: Counter[str] = Counter()
        self.id_counts: Counter[str] = Counter()
        self.parties: Counter[str] = Counter()
        self.active_legislators: defaultdict[
            str, defaultdict[str, list[Person]]
        ] = defaultdict(lambda: defaultdict(list))

    def summarize(self, person: Person) -> None:
        self.person_count += 1
        self.optional_fields.update(
            set(person.dict(exclude_defaults=True).keys()) & OPTIONAL_FIELD_SET
        )
        self.extra_counts.update(person.extras.keys())

        district = role_type = None
        for role in person.roles:
            if role.is_active():
                role_type = role.type
                district = role.district
                break
        if role_type and district:
            self.active_legislators[role_type][district].append(person)

        for p_role in person.party:
            if p_role.is_active():
                self.parties[p_role.name] += 1

        for cd in person.offices:
            for key in ("voice", "fax", "address"):
                if getattr(cd, key, None):
                    self.contact_counts[cd.classification + " " + key] += 1

        for scheme, id_ in person.ids or []:
            if id_:
                self.id_counts[scheme] += 1
        for id_ in person.other_identifiers:
            if id_.scheme not in ("openstates", "legacy_openstates"):
                self.id_counts[id_.scheme] += 1

    def print_summary(self) -> None:  # pragma: no cover
        click.secho(
            f"processed {self.person_count} active people",
            bold=True,
        )
        for role_type in self.active_legislators:
            count = sum([len(v) for v in self.active_legislators[role_type].values()])
            click.secho(f"{count:4d} {role_type}")

        click.secho("Parties", bold=True)
        for party, count in self.parties.items():
            if party == "Republican":
                color = "red"
            elif party == "Democratic":
                color = "blue"
            else:
                color = "green"
            click.secho(f"{count:4d} {party} ", bg=color)

        for name, collection in {
            "Contact Info": self.contact_counts,
            "Identifiers": self.id_counts,
            "Additional Info": self.optional_fields,
            "Extras": self.extra_counts,
        }.items():
            if collection:
                click.secho(name, bold=True)
                for type, count in collection.items():
                    click.secho(
                        f" {type:<25} {count:4d} {count/self.person_count*100:.0f}% "
                    )
            else:
                click.secho(name + " - none", bold=True)

    def print_roster(self) -> None:  # pragma: no cover
        for role_type, districts in self.active_legislators.items():
            for district, people in sorted(districts.items()):
                click.secho(f"{role_type} {district}", fg="blue")
                for person in people:
                    click.secho(f"   {person.name}")

    def process_legislature(self, abbr: str) -> None:  # pragma: no cover
        path = get_data_path(abbr) / "legislature"
        filenames = path.glob("*.yml")

        for filename in filenames:
            person: Person = Person.load_yaml(filename)
            self.summarize(person)


def write_csv(files: list[Path], jurisdiction_id: str, output_filename: str) -> None:
    with open(output_filename, "w") as outf:
        out = csv.DictWriter(
            outf,
            (
                "id",
                "name",
                "current_party",
                "current_district",
                "current_chamber",
                "given_name",
                "family_name",
                "gender",
                "email",
                "biography",
                "birth_date",
                "death_date",
                "image",
                "links",
                "sources",
                "capitol_address",
                "capitol_voice",
                "capitol_fax",
                "district_address",
                "district_voice",
                "district_fax",
                "twitter",
                "youtube",
                "instagram",
                "facebook",
            ),
        )
        out.writeheader()

        for filename in files:
            person: Person = Person.load_yaml(filename)

            # current party
            for p_role in person.party:
                if p_role.is_active():
                    current_party = p_role.name
                    break

            # current district
            for role in person.roles:
                if role.is_active():
                    current_chamber = role.type
                    current_district = role.district

            district_address = district_voice = district_fax = None
            capitol_address = capitol_voice = capitol_fax = None
            for cd in person.offices:
                if cd.classification == "district":
                    district_address = cd.address
                    district_voice = cd.voice
                    district_fax = cd.fax
                elif cd.classification == "district":
                    capitol_address = cd.address
                    capitol_voice = cd.voice
                    capitol_fax = cd.fax

            links = ";".join(k.url for k in person.links)
            sources = ";".join(k.url for k in person.sources)

            obj = {
                "id": person.id,
                "name": person.name,
                "current_party": current_party,
                "current_district": current_district,
                "current_chamber": current_chamber,
                "given_name": person.given_name,
                "family_name": person.family_name,
                "gender": person.gender,
                "email": person.email,
                "biography": person.biography,
                "birth_date": person.birth_date,
                "death_date": person.death_date,
                "image": person.image,
                "twitter": person.ids.twitter if person.ids else "",
                "youtube": person.ids.youtube if person.ids else "",
                "instagram": person.ids.instagram if person.ids else "",
                "facebook": person.ids.facebook if person.ids else "",
                "links": links,
                "sources": sources,
                "district_address": district_address,
                "district_voice": district_voice,
                "district_fax": district_fax,
                "capitol_address": capitol_address,
                "capitol_voice": capitol_voice,
                "capitol_fax": capitol_fax,
            }
            out.writerow(obj)

    click.secho(f"processed {len(files)} files", fg="green")


def lint_dir(
    abbr: str, verbose: bool, municipal: bool, date: str, fix: bool, save_all: bool
) -> int:  # pragma: no cover
    state_dir = get_data_path(abbr)
    legislative_filenames = (state_dir / "legislature").glob("*.yml")
    executive_filenames = (state_dir / "executive").glob("*.yml")
    municipality_filenames = (state_dir / "municipalities").glob("*.yml")
    retired_filenames = (state_dir / "retired").glob("*.yml")

    # a hack... need to get data path then traverse up
    settings_file = state_dir.parents[1] / "settings.yml"
    with open(settings_file) as f:
        settings = yaml.safe_load(f)

    try:
        validator = Validator(abbr, settings, fix, save_all)
    except BadVacancy:
        sys.exit(-1)

    all_filenames = [
        (PersonType.LEGISLATIVE, legislative_filenames),
        (PersonType.RETIRED, retired_filenames),
        (PersonType.EXECUTIVE, executive_filenames),
    ]

    if municipal:
        all_filenames.append((PersonType.MUNICIPAL, municipality_filenames))

    for person_type, filenames in all_filenames:
        for filename in filenames:
            # load just the data here since validate_person will convert to Person and
            # catch errors
            with open(filename) as file:
                data = yaml.safe_load(file)
            validator.validate_person(data, filename, person_type)

    error_count = validator.print_validation_report(verbose)

    return error_count


def _echo_org_status(org: typing.Any, created: bool, updated: bool) -> None:
    if created:
        click.secho(f"{org} created", fg="green")
    elif updated:
        click.secho(f"{org} updated", fg="yellow")


def load_directory_to_database(files: list[Path], purge: bool) -> None:
    from openstates.data.models import Person as DjangoPerson
    from openstates.data.models import BillSponsorship, PersonVote, Jurisdiction

    ids = set()
    merged = {}
    created_count = 0
    updated_count = 0

    all_data = []
    all_jurisdictions = []
    updated_jurisdictions = set()
    for filename in files:
        person: Person = Person.load_yaml(filename)
        all_data.append((person, filename))
        if person.roles:
            all_jurisdictions.append(person.roles[0].jurisdiction)

    existing_ids = set(
        DjangoPerson.objects.filter(
            memberships__organization__jurisdiction_id__in=all_jurisdictions
        ).values_list("id", flat=True)
    )

    for person, filename in all_data:
        ids.add(person.id)
        created, updated = load_person(person)

        if created:
            click.secho(f"created person from {filename}", fg="cyan", bold=True)
            created_count += 1
        elif updated:
            click.secho(f"updated person from {filename}", fg="cyan")
            updated_count += 1

        if created or updated:
            updated_jurisdictions.add(person.roles[0].jurisdiction)

    missing_ids = existing_ids - ids

    # check if missing ids are in need of a merge
    for missing_id in missing_ids:
        try:
            found = DjangoPerson.objects.get(
                identifiers__identifier=missing_id, identifiers__scheme="openstates"
            )
            merged[missing_id] = found.id
        except DjangoPerson.DoesNotExist:
            pass

    if merged:
        click.secho(f"{len(merged)} removed via merge", fg="yellow")
        for old, new in merged.items():
            click.secho(f"   {old} => {new}", fg="yellow")
            BillSponsorship.objects.filter(person_id=old).update(person_id=new)
            PersonVote.objects.filter(voter_id=old).update(voter_id=new)
            DjangoPerson.objects.filter(id=old).delete()
            missing_ids.remove(old)

    # ids that are still missing would need to be purged
    if missing_ids and not purge:
        click.secho(
            f"{len(missing_ids)} went missing, run with --purge to remove", fg="red"
        )
        for id in missing_ids:
            mobj = DjangoPerson.objects.get(pk=id)
            click.secho(f"  {id}: {mobj}")
        raise CancelTransaction()
    elif missing_ids and purge:
        click.secho(f"{len(missing_ids)} purged", fg="yellow")
        DjangoPerson.objects.filter(id__in=missing_ids).delete()

    if created_count or updated_count:
        Jurisdiction.objects.filter(id__in=updated_jurisdictions).update(
            latest_people_update=datetime.datetime.utcnow()
        )

    click.secho(
        f"processed {len(ids)} person files, {created_count} created, "
        f"{updated_count} updated",
        fg="green",
    )


def create_person(
    fname: str,
    lname: str,
    name: str,
    state: str,
    district: str,
    party: str,
    rtype: str,
    url: str,
    image: str,
    email: str,
    start_date: str,
) -> None:
    role = Role(
        type=rtype,
        district=district,
        jurisdiction=abbr_to_jid(state),
        start_date=start_date,
    )

    if rtype in ("upper", "lower", "legislature"):
        directory = "legislature"
    elif rtype in ("mayor",):
        directory = "municipalities"
    elif rtype in ("governor", "lt_governor"):
        directory = "executive"

    person = Person(
        id=ocd_uuid("person"),
        name=name or f"{fname} {lname}",
        given_name=fname,
        family_name=lname,
        image=image,
        email=email,
        party=[Party(name=party)],
        roles=[role],
        links=[Link(url=url)],
        sources=[Link(url=url)],
    )

    output_dir = get_data_path(state) / directory
    dump_obj(person, output_dir=output_dir)


@click.group()
def main() -> None:
    pass


@main.command()
@click.argument("abbreviations", nargs=-1)
@click.option(
    "--upload/--no-upload", default=False, help="Upload to S3. (default: false)"
)
def to_csv(abbreviations: list[str], upload: bool) -> None:
    """
    Generate CSV files for YAML and optionally sync to S3.
    """
    if not abbreviations:
        abbreviations = get_all_abbreviations()

    if upload:
        s3 = boto3.client("s3")

    for abbr in abbreviations:
        click.secho("==== {} ====".format(abbr), bold=True)
        jurisdiction_id = abbr_to_jid(abbr)
        directory = get_data_path(abbr)
        person_files = sorted((directory / "legislature").glob("*.yml"))
        fname = f"{abbr}.csv"
        write_csv(person_files, jurisdiction_id, fname)

        if upload:
            s3.upload_file(
                fname,
                "data.openstates.org",
                f"people/current/{abbr}.csv",
                ExtraArgs={
                    "ContentType": "text/csv; charset=UTF-8",
                    "ACL": "public-read",
                },
            )
            click.secho(
                f"uploaded to data.openstates.org/people/current/{abbr}.csv", fg="green"
            )


@main.command()
@click.option("--fname", prompt="First Name", help="First Name")
@click.option("--lname", prompt="Last Name", help="Last Name")
@click.option("--name", help="Optional Name, if not provided First + Last will be used")
@click.option("--state", prompt="State", help="State abbreviation")
@click.option("--district", prompt="District", help="District")
@click.option("--party", prompt="Party", help="Party")
@click.option("--rtype", prompt="Role Type (upper|lower|mayor)", help="Role Type")
@click.option("--url", prompt="URL", help="Source URL")
@click.option("--image", prompt="Image URL", help="Image URL")
@click.option("--email", prompt="Email", help="Email")
@click.option("--start-date", prompt="Start Date", help="Start Date YYYY-MM-DD")
def new(
    fname: str,
    lname: str,
    name: str,
    state: str,
    district: str,
    party: str,
    rtype: str,
    url: str,
    image: str,
    email: str,
    start_date: str,
) -> None:
    """
    Create a new person record.

    Arguments can be passed via command line flags, omitted arguments will be prompted.

    Be sure to review the file and add any additional data before committing.
    """
    create_person(
        fname=fname,
        lname=lname,
        name=name,
        state=state,
        district=district,
        party=party,
        rtype=rtype,
        url=url,
        image=image,
        email=email,
        start_date=start_date,
    )


@main.command()
@click.argument("abbreviations", nargs=-1)
@click.option("--roster/--no-roster", default=False, help="Print roster after summary.")
def summarize(abbreviations: list[str], roster: bool) -> None:
    """
    Provide summary of a jurisdiction's data.

    <ABBR> can be provided to restrict linting to single state's files.
    """
    if not abbreviations:
        abbreviations = get_all_abbreviations()

    summarizer = Summarizer()
    for abbr in abbreviations:
        summarizer.process_legislature(abbr)
    summarizer.print_summary()
    if roster:
        summarizer.print_roster()


@main.command()
@click.argument("filenames", nargs=-1)
@click.option("--date")
@click.option("--reason", default=None)
@click.option("--death", is_flag=True)
@click.option("--vacant", is_flag=True)
def retire(
    date: str,
    filenames: list[str],
    reason: typing.Optional[str],
    death: bool,
    vacant: bool,
) -> None:
    """
    Retire a legislator, given END_DATE and FILENAME.

    Will set end_date on active roles.
    """
    for filename in filenames:
        # end the person's active roles & re-save
        person: Person = Person.load_yaml(Path(filename))
        if death:
            reason = "Deceased"
        person, num = retire_person(person, date, reason, death)

        if vacant:
            # default to 60 days for now
            add_vacancy(
                person, until=datetime.datetime.today() + datetime.timedelta(days=60)
            )

        dump_obj(person, filename=filename)

        if num == 0:
            click.secho("no active roles to retire", fg="red")
        elif num == 1:
            click.secho("retired person")
        else:
            click.secho(f"retired person from {num} roles")

        new_filename = retire_file(filename)
        click.secho(f"moved from {filename} to {new_filename}")


@main.command()
@click.argument("abbreviations", nargs=-1)
@click.option(
    "--skip-existing/--no-skip-existing",
    help="Skip processing for files that already exist on S3. (default: true)",
)
def sync_images(abbreviations: list[str], skip_existing: bool) -> None:
    """
    Download images and sync them to S3.

    <ABBR> can be provided to restrict to single state.
    """
    if not abbreviations:
        abbreviations = get_all_abbreviations()

    for abbr in abbreviations:
        download_state_images(abbr, skip_existing)


@main.command()
@click.argument("abbreviations", nargs=-1)
@click.option("-v", "--verbose", count=True)
@click.option(
    "--fix/--no-fix", default=False, help="Enable/disable automatic fixing of data."
)
@click.option(
    "--save-all/--no-save-all",
    default=False,
    help="Enable/disable automatic reforamting of YAML.",
)
@click.option(
    "--municipal/--no-municipal",
    default=True,
    help="Enable/disable linting of municipal data.",
)
@click.option(
    "--date",
    type=str,
    default=None,
    help="Lint roles using a certain date instead of today.",
)
def lint(
    abbreviations: list[str],
    verbose: bool,
    municipal: bool,
    date: str,
    fix: bool,
    save_all: bool,
) -> None:
    """
    Lint YAML files.

    <ABBR> can be provided to restrict linting to single state's files.
    """
    error_count = 0

    if not abbreviations:
        abbreviations = get_all_abbreviations()

    for abbr in abbreviations:
        click.secho("==== {} ====".format(abbr), bold=True)
        error_count += lint_dir(abbr, verbose, municipal, date, fix, save_all)

    if error_count:
        click.secho(f"exiting with {error_count} errors", fg="red")
        sys.exit(99)


@main.command()
@click.argument("abbreviations", nargs=-1)
@click.option(
    "--purge/--no-purge",
    default=False,
    help="Purge all legislators from DB that aren't in YAML.",
)
@click.option(
    "--safe/--no-safe",
    default=False,
    help="Operate in safe mode, no changes will be written to database.",
)
def to_database(abbreviations: list[str], purge: bool, safe: bool) -> None:
    """
    Sync YAML files to DB.
    """
    init_django()

    create_parties()

    if not abbreviations:
        abbreviations = get_all_abbreviations()

    for abbr in abbreviations:
        click.secho("==== {} ====".format(abbr), bold=True)
        directory = get_data_path(abbr)
        municipalities = load_municipalities(abbr)

        with transaction.atomic():
            create_municipalities(municipalities)

        person_files = list(
            itertools.chain(
                directory.glob("legislature/*.yml"),
                directory.glob("executive/*.yml"),
                directory.glob("municipalities/*.yml"),
                directory.glob("retired/*.yml"),
            )
        )

        if safe:
            click.secho("running in safe mode, no changes will be made", fg="magenta")

        try:
            with transaction.atomic():
                load_directory_to_database(person_files, purge=purge)

                if safe:
                    click.secho("ran in safe mode, no changes were made", fg="magenta")
                    raise CancelTransaction()
        except CancelTransaction:
            sys.exit(1)


@main.command()  # pragma: no cover
@click.argument("abbr")
@click.argument("input_dir")
@click.option(
    "--retirement",
    default=None,
    help="Set retirement date for all people marked retired.",
)
def merge(abbr: str, input_dir: str, retirement: str) -> None:
    """
    Convert scraped JSON in INPUT_DIR to YAML files for this repo.
    """
    jurisdiction_id = abbr_to_jid(abbr)

    new_people = process_scrape_dir(Path(input_dir), jurisdiction_id)

    existing_people: list[Person] = []
    directory = get_data_path(abbr)
    for filename in itertools.chain(
        directory.glob("legislature/*.yml"),
        directory.glob("retired/*.yml"),
    ):
        existing_people.append(Person.load_yaml(filename))

    click.secho(
        f"analyzing {len(existing_people)} existing people and {len(new_people)} scraped"
    )

    unmatched = incoming_merge(abbr, existing_people, new_people, retirement)
    click.secho(f"{len(unmatched)} people were unmatched")


if __name__ == "__main__":
    main()



import os
from .utils import transformers

# scrape settings

SCRAPELIB_RPM = 60
SCRAPELIB_TIMEOUT = 60
SCRAPELIB_RETRY_ATTEMPTS = 3
SCRAPELIB_RETRY_WAIT_SECONDS = 10
SCRAPELIB_VERIFY = True

CACHE_DIR = os.path.join(os.getcwd(), "_cache")
SCRAPED_DATA_DIR = os.path.join(os.getcwd(), "_data")

IMPORT_TRANSFORMERS = {"bill": {"identifier": transformers.fix_bill_id}}

# Django settings
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "%(asctime)s %(levelname)s %(name)s: %(message)s",
            "datefmt": "%H:%M:%S",
        }
    },
    "handlers": {
        "default": {
            "level": "DEBUG",
            "class": "openstates.utils.ansistrm.ColorizingStreamHandler",
            "formatter": "standard",
        }
    },
    "loggers": {
        "": {"handlers": ["default"], "level": "DEBUG", "propagate": True},
        "scrapelib": {"handlers": ["default"], "level": "INFO", "propagate": False},
        "requests": {"handlers": ["default"], "level": "WARN", "propagate": False},
        "boto": {"handlers": ["default"], "level": "WARN", "propagate": False},
    },
}



from django.db import models
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.search import SearchVectorField
from django.contrib.postgres.indexes import GinIndex


from .base import (
    OCDBase,
    LinkBase,
    OCDIDField,
    RelatedBase,
    RelatedEntityBase,
    MimetypeLinkBase,
)
from .people_orgs import Organization
from .jurisdiction import LegislativeSession
from .. import common


class Bill(OCDBase):
    id = OCDIDField(ocd_type="bill")
    legislative_session = models.ForeignKey(
        LegislativeSession,
        related_name="bills",
        # sessions should be hard to delete
        on_delete=models.PROTECT,
    )
    identifier = models.CharField(max_length=100)

    title = models.TextField()

    from_organization = models.ForeignKey(
        Organization,
        related_name="bills",
        null=True,
        # chambers should be hard to delete
        on_delete=models.PROTECT,
    )
    # check that array values are in enum?
    classification = ArrayField(base_field=models.TextField(), blank=True, default=list)
    subject = ArrayField(base_field=models.TextField(), blank=True, default=list)

    # computed fields
    first_action_date = models.CharField(max_length=25, default=None, null=True)
    latest_action_date = models.CharField(max_length=25, default=None, null=True)
    latest_action_description = models.TextField(default="")
    latest_passage_date = models.CharField(max_length=25, default=None, null=True)

    def __str__(self):
        return "{} in {}".format(self.identifier, self.legislative_session)

    class Meta:
        db_table = "opencivicdata_bill"
        index_together = [["from_organization", "legislative_session", "identifier"]]


class BillAbstract(RelatedBase):
    bill = models.ForeignKey(Bill, related_name="abstracts", on_delete=models.CASCADE)
    abstract = models.TextField()
    note = models.TextField(blank=True)

    def __str__(self):
        return "{0} abstract".format(self.bill.identifier)

    class Meta:
        db_table = "opencivicdata_billabstract"


class BillTitle(RelatedBase):
    bill = models.ForeignKey(
        Bill, related_name="other_titles", on_delete=models.CASCADE
    )
    title = models.TextField()
    note = models.TextField(blank=True)

    def __str__(self):
        return "{0} ({1})".format(self.title, self.bill.identifier)

    class Meta:
        db_table = "opencivicdata_billtitle"


class BillIdentifier(RelatedBase):
    bill = models.ForeignKey(
        Bill, related_name="other_identifiers", on_delete=models.CASCADE
    )
    identifier = models.CharField(
        max_length=300,
        help_text="A unique identifier developed by an upstream or third party source.",
    )

    class Meta:
        db_table = "opencivicdata_billidentifier"


class BillAction(RelatedBase):
    bill = models.ForeignKey(Bill, related_name="actions", on_delete=models.CASCADE)
    organization = models.ForeignKey(
        Organization,
        related_name="actions",
        # don't let an org delete wipe out a bunch of bill actions
        on_delete=models.PROTECT,
    )
    description = models.TextField()
    date = models.CharField(max_length=25)  # YYYY-MM-DD HH:MM:SS+HH:MM
    classification = ArrayField(
        base_field=models.TextField(), blank=True, default=list
    )  # enum
    order = models.PositiveIntegerField()

    class Meta:
        db_table = "opencivicdata_billaction"
        ordering = ["order"]

    def __str__(self):
        return "{0} action on {1}".format(self.bill.identifier, self.date)


class BillActionRelatedEntity(RelatedEntityBase):
    action = models.ForeignKey(
        BillAction, related_name="related_entities", on_delete=models.CASCADE
    )

    def __str__(self):
        return "{0} related to {1}".format(self.entity_name, self.action)

    class Meta:
        db_table = "opencivicdata_billactionrelatedentity"


class RelatedBill(RelatedBase):
    bill = models.ForeignKey(
        Bill, related_name="related_bills", on_delete=models.CASCADE
    )
    related_bill = models.ForeignKey(
        Bill,
        related_name="related_bills_reverse",
        null=True,
        # if related bill goes away, just unlink the relationship
        on_delete=models.SET_NULL,
    )
    identifier = models.CharField(max_length=100)
    # not a FK in case we don't know the session yet
    legislative_session = models.CharField(max_length=100)
    relation_type = models.CharField(
        max_length=100, choices=common.BILL_RELATION_TYPE_CHOICES
    )

    def __str__(self):
        return "relationship of {} to {} ({})".format(
            self.bill, self.related_bill, self.relation_type
        )

    class Meta:
        db_table = "opencivicdata_relatedbill"


class BillSponsorship(RelatedEntityBase):
    bill = models.ForeignKey(
        Bill, related_name="sponsorships", on_delete=models.CASCADE
    )
    primary = models.BooleanField(default=False)
    classification = models.CharField(max_length=100)  # enum?

    def __str__(self):
        return "{} ({}) sponsorship of {}".format(
            self.name, self.entity_type, self.bill
        )

    class Meta:
        db_table = "opencivicdata_billsponsorship"


class BillDocument(RelatedBase):
    bill = models.ForeignKey(Bill, related_name="documents", on_delete=models.CASCADE)
    note = models.CharField(max_length=300)
    date = models.CharField(max_length=10)  # YYYY[-MM[-DD]]
    classification = models.CharField(
        max_length=100, choices=common.BILL_DOCUMENT_CHOICES, blank=True
    )
    extras = models.JSONField(default=dict, blank=True)

    def __str__(self):
        return "{0} document of {1}".format(self.date, self.bill)

    class Meta:
        db_table = "opencivicdata_billdocument"


class BillVersion(RelatedBase):
    bill = models.ForeignKey(Bill, related_name="versions", on_delete=models.CASCADE)
    note = models.CharField(max_length=300)
    date = models.CharField(max_length=10)  # YYYY[-MM[-DD]]
    classification = models.CharField(
        max_length=100, choices=common.BILL_VERSION_CHOICES, blank=True
    )
    extras = models.JSONField(default=dict, blank=True)

    def __str__(self):
        return "{0} version of {1}".format(self.date, self.bill)

    class Meta:
        db_table = "opencivicdata_billversion"


class BillDocumentLink(MimetypeLinkBase):
    document = models.ForeignKey(
        BillDocument, related_name="links", on_delete=models.CASCADE
    )

    def __str__(self):
        return "{0} for {1}".format(self.url, self.document.bill)

    class Meta:
        db_table = "opencivicdata_billdocumentlink"


class BillVersionLink(MimetypeLinkBase):
    version = models.ForeignKey(
        BillVersion, related_name="links", on_delete=models.CASCADE
    )

    def __str__(self):
        return "{0} for {1}".format(self.url, self.version)

    class Meta:
        db_table = "opencivicdata_billversionlink"


class BillSource(LinkBase):
    bill = models.ForeignKey(Bill, related_name="sources", on_delete=models.CASCADE)

    class Meta:
        db_table = "opencivicdata_billsource"


class SearchableBill(models.Model):
    """
    This model associates a single version's text with a given bill.

    This is done for a few reasons:
        * bills with multiple versions aren't weighted more heavily than others
        * this makes querying quite a bit more efficient (no need to deduplicate results)

    We'll also store error results, assuming that they're somewhat persistent.
    """

    bill = models.OneToOneField(
        Bill, related_name="searchable", null=True, on_delete=models.CASCADE
    )
    version_link = models.OneToOneField(
        BillVersionLink, related_name="searchable", null=True, on_delete=models.CASCADE
    )

    search_vector = SearchVectorField(default=None)
    all_titles = models.TextField(default="")
    raw_text = models.TextField(default="")
    is_error = models.BooleanField(default=False)

    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        db_table = "opencivicdata_searchablebill"
        indexes = [GinIndex(name="search_index", fields=["search_vector"])]



# Generated by Django 3.1.7 on 2021-04-01 14:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0030_auto_20210210_1817"),
    ]

    operations = [
        migrations.AddField(
            model_name="voteevent",
            name="dedupe_key",
            field=models.CharField(max_length=500, null=True),
        ),
        migrations.AlterField(
            model_name="jurisdiction",
            name="classification",
            field=models.CharField(
                choices=[
                    ("state", "State"),
                    ("country", "Country"),
                    ("municipality", "Municipality"),
                ],
                db_index=True,
                default="state",
                help_text="The type of Jurisdiction being defined.",
                max_length=50,
            ),
        ),
    ]



from openstates.utils import jid_to_abbr
from .common import (
    extract_simple_pdf,
    extract_line_numbered_pdf,
    extract_line_post_numbered_pdf,
    extract_pre_tag_html,
    extract_sometimes_numbered_pdf,
    extract_from_p_tags_html,
    extractor_for_elements_by_class,
    extractor_for_element_by_id,
    extractor_for_element_by_xpath,
    extract_from_code_tags_html,
    textract_extractor,
    Metadata,
    ExtractorFunc,
)
from .de import handle_delaware


class DoNotDownload:
    """ Sentinel to indicate that nothing should be downloaded """


DOCX_MIMETYPE = (
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
)

CONVERSION_FUNCTIONS = {
    "al": {"application/pdf": extract_line_numbered_pdf},
    "ak": {"text/html": extractor_for_element_by_id("draftOverlay")},
    "az": {
        "text/html": extractor_for_elements_by_class("WordSection2"),
        "application/pdf": DoNotDownload,
    },
    "ar": {"application/pdf": extract_sometimes_numbered_pdf},
    "ca": {
        "text/html": extractor_for_element_by_xpath(
            './/div[@id="bill"] | .//span[@class="Resolution"]'
        )
    },
    "co": {"application/pdf": extract_sometimes_numbered_pdf},
    "ct": {"text/html": extract_from_p_tags_html, "application/pdf": DoNotDownload},
    "dc": {"application/pdf": textract_extractor(extension="pdf", method="tesseract")},
    "de": {
        "text/html": handle_delaware,
        "application/pdf": handle_delaware,
        "application/msword": DoNotDownload,
    },
    "fl": {
        "text/html": extract_pre_tag_html,
        "application/pdf": extract_line_numbered_pdf,
    },
    "ga": {"application/pdf": extract_sometimes_numbered_pdf},
    "hi": {
        "text/html": extractor_for_element_by_xpath(
            './/*[@class="Section2"] | .//*[@class="WordSection2"]'
        ),
        "application/pdf": DoNotDownload,
    },
    "ia": {"application/pdf": extract_line_numbered_pdf, "text/html": DoNotDownload},
    "id": {"application/pdf": extract_line_numbered_pdf},
    "il": {"text/html": extract_from_code_tags_html},
    "in": {"application/pdf": extract_sometimes_numbered_pdf},
    "ks": {"application/pdf": extract_sometimes_numbered_pdf},
    "ky": {"application/pdf": extract_line_numbered_pdf},
    "la": {"application/pdf": extract_sometimes_numbered_pdf},
    "ma": {"application/pdf": extract_line_numbered_pdf},
    "md": {"application/pdf": extract_line_numbered_pdf},
    "me": {
        "text/html": extractor_for_elements_by_class("billtextbody"),
        "application/rtf": DoNotDownload,
        "application/pdf": DoNotDownload,
    },
    "mi": {"text/html": extractor_for_element_by_xpath('.//*[@class="WordSection1"]')},
    "mo": {"application/pdf": extract_line_numbered_pdf},
    "mn": {"text/html": extractor_for_element_by_id("document")},
    "ms": {
        "text/html": extractor_for_element_by_xpath('.//*[@class="WordSection1"]'),
        "application/pdf": extract_line_numbered_pdf,
    },
    "mt": {"application/pdf": extract_sometimes_numbered_pdf},
    "nc": {"application/pdf": extract_sometimes_numbered_pdf},
    "nd": {"application/pdf": extract_sometimes_numbered_pdf},
    "ne": {"application/pdf": extract_sometimes_numbered_pdf},
    "nh": {
        "application/pdf": extract_sometimes_numbered_pdf,
        "text/html": extract_from_p_tags_html,
    },
    "nj": {"text/html": extractor_for_element_by_xpath('.//*[@class="WordSection3"]')},
    # NY HTML is just summaries
    "ny": {
        "text/html": DoNotDownload,
        "application/pdf": extract_sometimes_numbered_pdf,
    },
    "nm": {
        "application/pdf": extract_sometimes_numbered_pdf,
        "text/html": DoNotDownload,
    },
    "nv": {"application/pdf": extract_sometimes_numbered_pdf},
    "oh": {"application/pdf": extract_line_post_numbered_pdf},
    "or": {"application/pdf": extract_sometimes_numbered_pdf},
    "ok": {"application/pdf": extract_sometimes_numbered_pdf},
    "sc": {"text/html": extract_from_p_tags_html},
    "sd": {"text/html": extractor_for_elements_by_class("fullContent")},
    "tn": {"application/pdf": extract_simple_pdf},
    "ut": {"application/pdf": extract_line_numbered_pdf},
    "pr": {
        "application/msword": textract_extractor(extension="doc"),
        DOCX_MIMETYPE: textract_extractor(extension="docx"),
    },
    "pa": {
        "application/msword": DoNotDownload,
        "text/html": DoNotDownload,
        "application/pdf": extract_line_numbered_pdf,
    },
    "ri": {"application/pdf": extract_sometimes_numbered_pdf},
    # aggressive, but the Washington & Texas HTML are both basically bare
    "tx": {"text/html": extractor_for_element_by_xpath("//html")},
    "va": {"text/html": extractor_for_element_by_id("mainC")},
    "vt": {"application/pdf": extract_sometimes_numbered_pdf},
    "wa": {"text/html": extractor_for_element_by_xpath("//html")},
    "wi": {
        "application/pdf": extract_sometimes_numbered_pdf,
        "text/html": DoNotDownload,
    },
    "wv": {"text/html": extractor_for_element_by_xpath('.//*[@class="textcontainer"]')},
    "wy": {"application/pdf": extract_sometimes_numbered_pdf},
}


def get_extract_func(metadata: Metadata) -> ExtractorFunc:
    try:
        state = jid_to_abbr(metadata["jurisdiction_id"])
        # ignore type here because DoNotDownload sentinels were in the way
        func = CONVERSION_FUNCTIONS[state][metadata["media_type"]]  # type: ignore
    except KeyError:
        print(f"no function for {state}, {metadata['media_type']}")
        return lambda data, metadata: ""
    return func



# type: ignore
import re

from ..data import STATES_AND_DC_PR
from ..data.states import NC, VT


def test_basics() -> None:
    unicam_count = 0
    bicam_count = 0
    for state in STATES_AND_DC_PR:
        if state.unicameral:
            unicam_count += 1
            assert state.legislature
            assert not state.lower
            assert not state.upper
        else:
            bicam_count += 1
            assert not state.legislature
            assert state.lower
            assert state.upper
    assert unicam_count == 2
    assert bicam_count == 50


def test_district_numbers() -> None:
    for state in STATES_AND_DC_PR:
        if state.unicameral:
            assert state.legislature.num_seats == sum(
                d.num_seats for d in state.legislature.districts
            )
            assert state.legislature.num_seats > 0
        else:
            assert state.upper.num_seats == sum(
                d.num_seats for d in state.upper.districts
            )
            assert state.lower.num_seats == sum(
                d.num_seats for d in state.lower.districts
            )
            assert state.lower.num_seats > state.upper.num_seats > 10


def test_simple_numbered_districts() -> None:
    assert NC.lower.districts[0].name == "1"
    assert NC.lower.districts[0].num_seats == 1
    assert NC.lower.districts[-1].name == "120"
    assert NC.lower.districts[0].num_seats == 1


def test_everything_has_division_id() -> None:
    for state in STATES_AND_DC_PR:
        if state.unicameral:
            for d in state.legislature.districts:
                assert d.num_seats
                assert d.division_id.startswith(state.division_id)
        elif state.abbr == "US":
            for d in state.lower.districts:
                assert d.num_seats == 1
                assert re.match(
                    r"ocd-division/country:us/(district:dc|territory:(pr|as|mp|gu|vi)|state:\w\w/cd:)",
                    d.division_id,
                )
            for d in state.upper.districts:
                assert d.num_seats == 2
                assert re.match(r"ocd-division/country:us/state:\w\w$", d.division_id)
        else:
            for d in state.lower.districts:
                assert d.num_seats
                assert d.division_id.startswith(f"{state.division_id}/sldl:")
            for d in state.upper.districts:
                assert d.num_seats
                if d.name == "At-Large":
                    assert d.division_id == state.division_id
                else:
                    assert d.division_id.startswith(f"{state.division_id}/sldu:")


def test_everything_has_organization_id() -> None:
    for state in STATES_AND_DC_PR:
        if state.unicameral:
            assert len(state.legislature.organization_id) == 53
        else:
            assert len(state.lower.organization_id) == 53
            assert len(state.upper.organization_id) == 53


def test_legacy_ids() -> None:
    assert not NC.legacy_districts
    assert len(VT.legacy_districts) == 38



from ..models import State, Chamber, simple_numbered_districts

CO = State(
    name="Colorado",
    abbr="CO",
    capital="Denver",
    capital_tz="America/Denver",
    fips="08",
    unicameral=False,
    legislature_name="Colorado General Assembly",
    legislature_organization_id="ocd-organization/2fc73234-69ce-4a42-b85a-4a52a676e6e9",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/1f6f6532-1bc8-4ddb-8350-27fcd47c6795",
    division_id="ocd-division/country:us/state:co",
    jurisdiction_id="ocd-jurisdiction/country:us/state:co/government",
    url="http://leg.colorado.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/e047d0e5-2c93-4e39-bea3-04c0d9d11b20",
        num_seats=65,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:co", "lower", 65
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/98d43d46-9571-4a40-9007-581d84d41bb8",
        num_seats=35,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:co", "upper", 35
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

KY = State(
    name="Kentucky",
    abbr="KY",
    capital="Frankfort",
    capital_tz="America/New_York",
    fips="21",
    unicameral=False,
    legislature_name="Kentucky General Assembly",
    legislature_organization_id="ocd-organization/b5a82ff4-4fd3-41c8-bb44-32ad2e499482",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/d98917f5-3395-595e-b2d9-873936707342",
    division_id="ocd-division/country:us/state:ky",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ky/government",
    url="http://www.lrc.ky.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/4e2398aa-4ffd-4e2c-8089-7a0cb64f1bc9",
        num_seats=100,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ky", "lower", 100
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/523aa254-9d04-4641-9770-6488b5f4fb0e",
        num_seats=38,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ky", "upper", 38
        ),
    ),
)



import os
from pathlib import Path
import pytest  # type: ignore
from openstates.utils.people.merge import (
    reformat_phone_number,
    reformat_address,
    compute_merge,
    Append,
    Replace,
    merge_people,
    merge_offices,
    find_file,
)
from openstates.models.people import OtherName, OtherIdentifier, Office
from pydantic import BaseModel


@pytest.mark.parametrize(
    "input,output",
    [
        ("1234567890", "123-456-7890"),
        ("123-456-7890", "123-456-7890"),
        ("1-123-456-7890", "1-123-456-7890"),
        ("+1-123-456-7890", "1-123-456-7890"),
        ("1-800-FAKENUM", "1-800-FAKENUM"),
        ("email@example.com", "email@example.com"),
        ("555.333.1111", "555-333-1111"),
        ("+1 (555) 333-1111", "1-555-333-1111"),
        ("555-333-1111 ext.100", "555-333-1111 ext. 100"),
        ("555.333.1111 EXT.100", "555-333-1111 ext. 100"),
    ],
)
def test_reformat_phone(input, output):
    assert reformat_phone_number(input) == output


@pytest.mark.parametrize(
    "input,output",
    [
        ("123 Maple Lane\nRaleigh, NC 27511", "123 Maple Lane;Raleigh, NC 27511"),
        (
            "123 Maple Lane   \n   Raleigh, NC    27511",
            "123 Maple Lane;Raleigh, NC 27511",
        ),
        ("123 Maple Lane\n \nRaleigh, NC 27511", "123 Maple Lane;Raleigh, NC 27511"),
    ],
)
def test_reformat_address(input, output):
    assert reformat_address(input) == output


@pytest.mark.parametrize(
    "a, b, output",
    [
        # no diff
        ({"a": "one", "b": "two"}, {"a": "one", "b": "two"}, []),
        # simple difference
        (
            {"a": "one", "b": "two"},
            {"a": "bad", "b": "two"},
            [Replace("a", "one", "bad")],
        ),
        # missing key from first
        ({"a": "one"}, {"a": "one", "b": "two"}, [Replace("b", None, "two")]),
        # missing key from second
        ({"a": "one", "b": "two"}, {"a": "one"}, []),
    ],
)
def test_compute_merge_simple(a, b, output):
    class Model(BaseModel):
        a: str = None
        b: str = None

    assert compute_merge(Model(**a), Model(**b)) == output


@pytest.mark.parametrize(
    "a, b, output",
    [
        # no diff
        ({"a": {"b": "c"}}, {"a": {"b": "c"}}, []),
        # nothing new on right
        ({"a": {"b": "c"}}, {"a": {}}, []),
        # replace
        ({"a": {}}, {"a": {"b": "c"}}, [Replace("a.b", None, "c")]),
    ],
)
def test_compute_merge_nested(a, b, output):
    class Inner(BaseModel):
        b: str = None

    class Model(BaseModel):
        a: Inner = None

    assert compute_merge(Model(**a), Model(**b)) == output


@pytest.mark.parametrize(
    "a, b, output",
    [
        # no errors
        ({"a": [{"b": 1}, {"c": 2}]}, {"a": [{"b": 1}, {"c": 2}]}, []),
        # no errors - just different order
        ({"a": [{"b": 1}, {"c": 2}]}, {"a": [{"c": 2}, {"b": 1}]}, []),
        # extra item left
        ({"a": [{"b": 1}, {"c": 2}, {"d": 3}]}, {"a": [{"b": 1}, {"c": 2}]}, []),
        # extra item right
        (
            {"a": [{"b": 1}, {"c": 2}]},
            {"a": [{"b": 1}, {"c": 2}, {"d": 3}]},
            [Append("a", {"d": 3})],
        ),
    ],
)
def test_compute_merge_list(a, b, output):
    class Model(BaseModel):
        a: list[dict]

    assert compute_merge(Model(**a), Model(**b)) == output


@pytest.mark.parametrize(
    "a, b, keep_both, output",
    [
        # discard id
        ({"id": "1"}, {"id": "2"}, False, []),
        # keep id
        (
            {"id": "1"},
            {"id": "2"},
            True,
            [
                Append(
                    "other_identifiers",
                    OtherIdentifier(identifier="2", scheme="openstates"),
                )
            ],
        ),
        # append name
        (
            {"name": "A"},
            {"name": "B"},
            True,
            [Append("other_names", OtherName(name="A")), Replace("name", "A", "B")],
        ),
    ],
)
def test_compute_merge_special_cases(a, b, keep_both, output):
    class Model(BaseModel):
        id: str = None
        name: str = ""

    assert compute_merge(Model(**a), Model(**b), keep_both_ids=keep_both) == output


@pytest.mark.parametrize(
    "old, new, expected",
    [
        # no changes
        ({"name": "Anna"}, {"name": "Anna"}, {"name": "Anna"}),
        # field only in old
        (
            {"name": "Anna", "birth_date": "1980"},
            {"name": "Anna"},
            {"name": "Anna", "birth_date": "1980"},
        ),
        # field only in new
        (
            {"name": "Anna"},
            {"name": "Anna", "birth_date": "1980"},
            {"name": "Anna", "birth_date": "1980"},
        ),
        # special: name field differs
        (
            {"name": "Bob"},
            {"name": "Robert"},
            {"name": "Robert", "other_names": [OtherName(name="Bob")]},
        ),
    ],
)
def test_simple_merge(old, new, expected):
    class Model(BaseModel):
        name: str
        birth_date: str = ""
        other_names: list[OtherName] = []

    assert merge_people(Model(**old), Model(**new)) == Model(**expected)


@pytest.mark.parametrize(
    "old, new, expected",
    [
        # more in first list
        (
            {"other_names": [{"name": "A"}, {"name": "B"}]},
            {"other_names": [{"name": "A"}]},
            {"other_names": [{"name": "A"}, {"name": "B"}]},
        ),
        # more in second list
        (
            {"other_names": [{"name": "A"}]},
            {"other_names": [{"name": "A"}, {"name": "B"}]},
            {"other_names": [{"name": "A"}, {"name": "B"}]},
        ),
        # each list is unique
        (
            {"other_names": [{"name": "A"}]},
            {"other_names": [{"name": "B"}]},
            {"other_names": [{"name": "A"}, {"name": "B"}]},
        ),
    ],
)
def test_list_merge(old, new, expected):
    class Model(BaseModel):
        other_names: list[OtherName] = []

    # note that keep doesn't matter for these
    assert merge_people(Model(**old), Model(**new), None) == Model(**expected)


@pytest.mark.parametrize(
    "old, new, expected",
    [
        # simplest case
        (
            {"id": "ocd-person/1"},
            {"id": "ocd-person/2"},
            {
                "id": "ocd-person/1",
                "other_identifiers": [
                    OtherIdentifier(scheme="openstates", identifier="ocd-person/2")
                ],
            },
        ),
        # already has identifiers
        (
            {
                "id": "ocd-person/1",
                "other_identifiers": [
                    OtherIdentifier(scheme="openstates", identifier="ocd-person/0")
                ],
            },
            {"id": "ocd-person/2"},
            {
                "id": "ocd-person/1",
                "other_identifiers": [
                    OtherIdentifier(scheme="openstates", identifier="ocd-person/0"),
                    OtherIdentifier(scheme="openstates", identifier="ocd-person/2"),
                ],
            },
        ),
    ],
)
def test_keep_both_ids(old, new, expected):
    class Model(BaseModel):
        id: str
        other_identifiers: list[OtherIdentifier] = []

    assert merge_people(Model(**old), Model(**new), keep_both_ids=True) == Model(
        **expected
    )


@pytest.mark.parametrize(
    "old, new",
    [
        (
            [Office(classification="capitol", voice="123-555-9999")],
            [Office(classification="capitol", voice="123-555-9999")],
        ),
        (
            [
                Office(classification="capitol", voice="123-555-9999"),
                Office(classification="district", address="abc"),
            ],
            [
                Office(classification="capitol", voice="123-555-9999"),
                Office(classification="district", address="abc"),
            ],
        ),
    ],
)
def test_merge_office_no_change(old, new):
    assert merge_offices(old, new) is None


@pytest.mark.parametrize(
    "old, new, expected",
    [
        # replace a value with a new one
        (
            [Office(classification="capitol", voice="111-111-1111")],
            [Office(classification="capitol", voice="222-222-2222")],
            [Office(classification="capitol", voice="222-222-2222")],
        ),
        # merge two partial records
        (
            [Office(classification="capitol", voice="111-111-1111")],
            [Office(classification="capitol", fax="444-444-4444")],
            [
                Office(
                    classification="capitol", voice="111-111-1111", fax="444-444-4444"
                )
            ],
        ),
        # merge two offices into a single list
        (
            [Office(classification="capitol", voice="111-111-1111")],
            [Office(classification="district", voice="444-444-4444")],
            [
                Office(classification="capitol", voice="111-111-1111"),
                Office(classification="district", voice="444-444-4444"),
            ],
        ),
    ],
)
def test_merge_offices_changes(old, new, expected):
    assert merge_offices(old, new) == expected


def test_merge_extras():
    # replace was adding a key like extras._internal_id
    old = {"extras": {"_internal_id": 123}}
    new = {}
    expected = old.copy()

    class Model(BaseModel):
        extras: dict = {}

    assert merge_people(Model(**new), Model(**old)) == expected


def test_find_file_good():
    os.environ["OS_PEOPLE_DIRECTORY"] = str(Path(__file__).parent / "testdata/people")
    filename = find_file("a2e4a1b2-f0fd-4c35-9e0c-bb009778792f", state="pa")
    assert "Pam-Snyder" in str(filename)
    filename = find_file("a2e4a1b2-f0fd-4c35-9e0c-bb009778792f")
    assert "Pam-Snyder" in str(filename)


def test_find_file_missing():
    with pytest.raises(FileNotFoundError):
        find_file("77777777-ffff-0000-9000-bbbbbbbbbbbb")
    # good file, bad directory
    with pytest.raises(FileNotFoundError):
        find_file("a2e4a1b2-f0fd-4c35-9e0c-bb009778792f", state="nc")



from ..models import State, Chamber, simple_numbered_districts

LA = State(
    name="Louisiana",
    abbr="LA",
    capital="Baton Rouge",
    capital_tz="America/Chicago",
    fips="22",
    unicameral=False,
    legislature_name="Louisiana Legislature",
    legislature_organization_id="ocd-organization/3c509bec-a8f0-4422-a17f-5f11bfe88f26",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/6478a11e-ce20-5ee2-ab1c-055aae53cfcf",
    division_id="ocd-division/country:us/state:la",
    jurisdiction_id="ocd-jurisdiction/country:us/state:la/government",
    url="http://www.legis.la.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/d711b9b2-6668-470d-83a7-92d73b24f94a",
        num_seats=105,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:la", "lower", 105
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/ebc2f274-ac99-40a5-ae55-19e896936bc7",
        num_seats=39,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:la", "upper", 39
        ),
    ),
)



# Generated by Django 3.0.5 on 2020-09-14 10:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0021_remove_jurisdiction_feature_flags"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="organization",
            name="dissolution_date",
        ),
        migrations.RemoveField(
            model_name="organization",
            name="founding_date",
        ),
        migrations.RemoveField(
            model_name="organization",
            name="image",
        ),
    ]



# Generated by Django 3.0.5 on 2020-10-22 16:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0027_auto_20200923_1329"),
    ]

    operations = [
        migrations.AddField(
            model_name="person",
            name="email",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The official email address of the Person.",
                max_length=300,
            ),
        ),
        migrations.AlterField(
            model_name="personcontactdetail",
            name="value",
            field=models.CharField(
                help_text="The content of the Contact information like a phone number or address.",
                max_length=300,
            ),
        ),
    ]



"""
functions in this module should take a single object and update it in place

optionally, they can take a save parameter, that should default to False
but can be set to True to force a save if changes were made
(this allows for usage from CLI)
"""
from ._types import Model


def update_bill_fields(bill: Model, *, save: bool = False) -> None:
    first_action_date = None
    latest_action_date = None
    latest_action_description = ""
    latest_passage_date = None

    # iterate over according to order
    # first action date will use first by order (<)
    # latest will use latest by order (>=)
    for action in bill.actions.order_by("order"):
        if not first_action_date or action.date < first_action_date:
            first_action_date = action.date
        if not latest_action_date or action.date >= latest_action_date:
            latest_action_date = action.date
            latest_action_description = action.description
        if "passage" in action.classification and (
            not latest_passage_date or action.date >= latest_passage_date
        ):
            latest_passage_date = action.date

    if (
        bill.first_action_date != first_action_date
        or bill.latest_action_date != latest_action_date
        or bill.latest_passage_date != latest_passage_date
        or bill.latest_action_description != latest_action_description
    ):
        bill.first_action_date = first_action_date
        bill.latest_passage_date = latest_passage_date
        bill.latest_action_date = latest_action_date
        bill.latest_action_description = latest_action_description
        bill.save()



import pytest  # type: ignore
from openstates.cli.update import override_settings


class _Settings:
    pass


@pytest.fixture
def settings():
    ret = _Settings()
    ret.foo = "bar"
    ret.baz = "bob"
    return ret


def test_override_settings(settings):
    with override_settings(settings, {"baz": "fez"}):
        assert settings.foo == "bar"
        assert settings.baz == "fez"
    assert settings.foo == "bar"
    assert settings.baz == "bob"


def test_override_settings_unset(settings):
    with override_settings(settings, {"qux": "fez"}):
        assert settings.qux == "fez"
    assert not hasattr(settings, "qux")



import typing
import datetime
from functools import lru_cache
import click
from openstates import metadata
from .general import legacy_districts
from ...models.people import MAJOR_PARTIES, Person, PartyName


DataDict = dict[str, typing.Any]
# TODO: whenever Django typed replace these
DjangoModel = typing.Any
DjangoModelInstance = typing.Any


class CancelTransaction(Exception):
    pass


@lru_cache(128)
def cached_lookup(ModelCls: DjangoModel, **kwargs: str) -> DjangoModelInstance:
    # return ModelCls.objects.get(**kwargs)
    m = ModelCls.objects.get(**kwargs)
    return m


def update_subobjects(
    person: DjangoModelInstance,
    fieldname: str,
    objects: list[DataDict],
    read_manager: typing.Any = None,
) -> bool:
    """ returns True if there are any updates """
    # we need the default manager for this field in case we need to do updates
    manager = getattr(person, fieldname)

    # if a read_manager is passed, we'll use that for all read operations
    # this is used for DjangoPerson.memberships to ensure we don't wipe out committee memberships
    if read_manager is None:
        read_manager = manager

    current_count = read_manager.count()
    updated = False

    # if counts differ, we need to do an update for sure
    if current_count != len(objects):
        updated = True

    # check if all objects exist
    if not updated:
        qs = read_manager
        for obj in objects:
            qs = qs.exclude(**obj)

        if qs.exists():
            updated = True

    # if there's been an update, wipe the old & insert the new
    if updated:
        if current_count:
            read_manager.all().delete()
        for obj in objects:
            manager.create(**obj)
        # save to bump updated_at timestamp
        person.save()

    return updated


def get_update_or_create(
    ModelCls: DjangoModel, data: dict, lookup_keys: list[str]
) -> tuple[DjangoModelInstance, bool, bool]:
    updated = created = False
    kwargs = {k: data[k] for k in lookup_keys}
    try:
        obj = ModelCls.objects.get(**kwargs)
        for field, value in data.items():
            # special case datetime since comparisons won't work between str/datetime
            if isinstance(value, datetime.date):
                value = str(value)
            if getattr(obj, field) != value:
                setattr(obj, field, value)
                updated = True
        if updated:
            obj.save()
    except ModelCls.DoesNotExist:
        obj = ModelCls.objects.create(**data)
        created = True
    return obj, created, updated


def load_person(data: Person) -> tuple[bool, bool]:
    # import has to be here so that Django is set up
    from openstates.data.models import Organization, Post
    from openstates.data.models import Person as DjangoPerson

    fields = dict(
        id=data.id,
        name=data.name,
        given_name=data.given_name,
        family_name=data.family_name,
        gender=data.gender,
        email=data.email,
        biography=data.biography,
        birth_date=data.birth_date,
        death_date=data.death_date,
        image=data.image,
        extras=data.extras,
    )
    person, created, updated = get_update_or_create(DjangoPerson, fields, ["id"])

    updated |= update_subobjects(
        person, "other_names", [n.dict() for n in data.other_names]
    )
    updated |= update_subobjects(person, "links", [n.dict() for n in data.links])
    updated |= update_subobjects(person, "sources", [n.dict() for n in data.sources])
    updated |= update_subobjects(person, "offices", [n.dict() for n in data.offices])

    identifiers = []
    for scheme, value in data.ids.dict().items():
        if value:
            identifiers.append({"scheme": scheme, "identifier": value})
    for identifier in data.other_identifiers:
        identifiers.append(
            {"scheme": identifier.scheme, "identifier": identifier.identifier}
        )
    updated |= update_subobjects(person, "identifiers", identifiers)

    memberships = []
    primary_party = ""
    current_jurisdiction_id = None
    current_role = None
    for party in data.party:
        party_name = party.name
        try:
            org = cached_lookup(Organization, classification="party", name=party.name)
        except Organization.DoesNotExist:
            click.secho(f"no such party {party.name}", fg="red")
            raise CancelTransaction()
        memberships.append(
            {
                "organization": org,
                "start_date": party.start_date,
                "end_date": party.end_date,
            }
        )
        if party.is_active():
            if primary_party in MAJOR_PARTIES and party_name in MAJOR_PARTIES:
                raise ValueError(f"two primary parties for ({data.name} {data.id})")
            elif primary_party in MAJOR_PARTIES:
                # already set correct primary party, so do nothing
                pass
            else:
                primary_party = party_name
    for role in data.roles:
        if role.type == "mayor":
            role_name = "Mayor"
            org_type = "government"
            use_district = False
        elif role.type == "governor":
            role_name = "Governor"
            if (
                role.jurisdiction
                == "ocd-jurisdiction/country:us/district:dc/government"
            ):
                role_name = "Mayor"
            org_type = "executive"
            use_district = False
        elif role.type in ("secretary of state", "chief election officer"):
            role_name = role.type.title()
            org_type = "executive"
            use_district = False
        elif role.type in ("upper", "lower", "legislature"):
            org_type = role.type
            use_district = True
        else:
            raise ValueError(f"unsupported role type: {role.type}")
        try:
            org = cached_lookup(
                Organization, classification=org_type, jurisdiction_id=role.jurisdiction
            )
            if use_district:
                post = org.posts.get(label=role.district)
            else:
                post = None
        except Organization.DoesNotExist:
            click.secho(
                f"{person} no such organization {role.jurisdiction} {org_type}",
                fg="red",
            )
            raise CancelTransaction()
        except Post.DoesNotExist:
            # if this is a legacy district, be quiet
            lds = legacy_districts(jurisdiction_id=role.jurisdiction)
            if role.district not in lds[role.type]:
                click.secho(f"no such post {role}", fg="red")
                raise CancelTransaction()
            else:
                post = None

        if role.is_active():
            current_jurisdiction_id = role.jurisdiction

            current_role = {
                "org_classification": org_type,
                "district": None,
                "division_id": None,
            }
            if use_district:
                state_metadata = metadata.lookup(jurisdiction_id=role.jurisdiction)
                district = state_metadata.lookup_district(
                    name=str(role.district), chamber=role.type
                )
                assert district
                current_role["division_id"] = district.division_id
                current_role["title"] = getattr(state_metadata, role.type).title
                # try to force district to an int for sorting, but allow strings for non-numeric districts
                try:
                    current_role["district"] = int(role.district)  # type: ignore
                except ValueError:
                    current_role["district"] = str(role.district)
            else:
                current_role["title"] = role_name
        elif not current_jurisdiction_id:
            current_jurisdiction_id = role.jurisdiction

        membership = {
            "organization": org,
            "post": post,
            "start_date": role.start_date,
            "end_date": role.end_date,
        }
        if not use_district:
            membership["role"] = role_name
        memberships.append(membership)

    # note that we don't manage committee memberships here
    updated |= update_subobjects(
        person,
        "memberships",
        memberships,
        read_manager=person.memberships.exclude(
            organization__classification="committee"
        ),
    )

    # set computed fields (avoid extra save)
    if (
        person.primary_party != primary_party
        or person.current_role != current_role
        or person.current_jurisdiction_id != current_jurisdiction_id
    ):
        person.primary_party = primary_party
        person.current_role = current_role
        person.current_jurisdiction_id = current_jurisdiction_id
        person.save()

    return created, updated


def create_parties() -> None:
    from openstates.data.models import Organization

    for party in PartyName:
        org, created = Organization.objects.get_or_create(
            name=party.value, classification="party"
        )
        if created:
            click.secho(f"created party: {party.value}", fg="green")


def create_municipalities(jurisdictions: list[DataDict]) -> None:
    from openstates.data.models import Jurisdiction, Organization

    for jurisdiction in jurisdictions:
        j, created = Jurisdiction.objects.get_or_create(
            id=jurisdiction["id"],
            name=jurisdiction["name"],
            classification="municipality",
        )
        if created:
            click.secho(f"created jurisdiction: {j.name}", fg="green")

        o, created = Organization.objects.get_or_create(
            jurisdiction=j,
            classification="government",
            name=f"{jurisdiction['name']} Government",
        )
        if created:
            click.secho(f"created organization: {o.name}", fg="green")



import typing
from django.db.models import Model, Q

# type aliases
_ID = typing.Union[int, str]
_JsonDict = typing.Dict[str, typing.Any]
_DBSpec = typing.Union[_JsonDict, Q]
# TODO: the third parameter in the tuple should be _RelatedModels, but this doesn't
# play nicely with mypy (nested dictionary types may not be supported?)
_RelatedModels = typing.Dict[str, typing.Tuple[Model, str, typing.Dict]]
_TransformerFunc = typing.Callable[[str], str]
_TransformerMapping = typing.Dict[str, _TransformerFunc]



from ..models import District

legacy_districts = {
    "md": [
        District("12A", "lower", division_id=None),
        District("12B", "lower", division_id=None),
        District("2C", "lower", division_id=None),
        District("30", "lower", division_id=None),
        District("31", "lower", division_id=None),
        District("33A", "lower", division_id=None),
        District("33B", "lower", division_id=None),
        District("42", "lower", division_id=None),
        District("44", "lower", division_id=None),
        District("47", "lower", division_id=None),
        District("4A", "lower", division_id=None),
        District("4B", "lower", division_id=None),
        District("5A", "lower", division_id=None),
        District("5B", "lower", division_id=None),
    ],
    "nv": [
        District("Capital Senatorial District", "upper", division_id=None),
        District("Central Nevada Senatorial District", "upper", division_id=None),
        District("Clark County, No. 1", "upper", division_id=None),
        District("Clark County, No. 3", "upper", division_id=None),
        District("Clark County, No. 4", "upper", division_id=None),
        District("Clark County, No. 5", "upper", division_id=None),
        District("Clark County, No. 6", "upper", division_id=None),
        District("Clark County, No. 7", "upper", division_id=None),
        District("Clark County, No. 8", "upper", division_id=None),
        District("Clark County, No. 9", "upper", division_id=None),
        District("Clark County, No. 9", "upper", division_id=None),
        District("Clark County, No. 10", "upper", division_id=None),
        District("Clark County, No. 11", "upper", division_id=None),
        District("Clark County, No. 12", "upper", division_id=None),
        District("Rural Nevada Senatorial District", "upper", division_id=None),
        District("Washoe County, No. 1", "upper", division_id=None),
        District("Washoe County, No. 2", "upper", division_id=None),
        District("Washoe County, No. 3", "upper", division_id=None),
        District("Washoe County, No. 4", "upper", division_id=None),
    ],
    "pr": [District(str(n), "upper", division_id=None) for n in range(1, 9)],
    "vt": [
        District("Addison-Rutland-1", "lower", division_id=None),
        District("Bennington-Rutland-1", "lower", division_id=None),
        District("Chittenden-1-2", "lower", division_id=None),
        District("Chittenden-3-1", "lower", division_id=None),
        District("Chittenden-3-2", "lower", division_id=None),
        District("Chittenden-3-3", "lower", division_id=None),
        District("Chittenden-3-4", "lower", division_id=None),
        District("Chittenden-3-5", "lower", division_id=None),
        District("Chittenden-3-6", "lower", division_id=None),
        District("Chittenden-3-7", "lower", division_id=None),
        District("Chittenden-3-9", "lower", division_id=None),
        District("Chittenden-4", "lower", division_id=None),
        District("Chittenden-8", "lower", division_id=None),
        District("Chittenden-9", "lower", division_id=None),
        District("Franklin-3", "lower", division_id=None),
        District("Grand Isle-Chittenden-1-1", "lower", division_id=None),
        District("Lamoille-4", "lower", division_id=None),
        District("Lamoille-Washington-1", "lower", division_id=None),
        District("Orange-Addison-1", "lower", division_id=None),
        District("Orleans-Caledonia-1", "lower", division_id=None),
        District("Rutland-1-1", "lower", division_id=None),
        District("Rutland-1-2", "lower", division_id=None),
        District("Rutland-7", "lower", division_id=None),
        District("Rutland-8", "lower", division_id=None),
        District("Washington-3-2", "lower", division_id=None),
        District("Washington-3-3", "lower", division_id=None),
        District("Washington-Chittenden-1", "lower", division_id=None),
        District("Windham-2", "lower", division_id=None),
        District("Windham-3-1", "lower", division_id=None),
        District("Windham-3-3", "lower", division_id=None),
        District("Windham-Bennington-1", "lower", division_id=None),
        District("Windham-Bennington-Windsor-1", "lower", division_id=None),
        District("Windsor-1-1", "lower", division_id=None),
        District("Windsor-1-2", "lower", division_id=None),
        District("Windsor-3", "lower", division_id=None),
        District("Windsor-4", "lower", division_id=None),
        District("Windsor-6-1", "lower", division_id=None),
        District("Windsor-6-2", "lower", division_id=None),
    ],
    "us": [
        District("IL-20", "lower", division_id=None),
        District("MA-10", "lower", division_id=None),
        District("MO-9", "lower", division_id=None),
        District("NJ-13", "lower", division_id=None),
        District("NY-29", "lower", division_id=None),
        District("OH-17", "lower", division_id=None),
        District("OH-18", "lower", division_id=None),
        District("OK-6", "lower", division_id=None),
    ],
}



from ..models import State, Chamber, simple_numbered_districts

PA = State(
    name="Pennsylvania",
    abbr="PA",
    capital="Harrisburg",
    capital_tz="America/New_York",
    fips="42",
    unicameral=False,
    legislature_name="Pennsylvania General Assembly",
    legislature_organization_id="ocd-organization/4a283791-567b-48b8-aac3-c94c36e3e63a",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/ea33e6ed-4944-5e83-888b-c98b070907ce",
    division_id="ocd-division/country:us/state:pa",
    jurisdiction_id="ocd-jurisdiction/country:us/state:pa/government",
    url="http://www.legis.state.pa.us/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/0f78429d-ec86-4333-9338-17708fe11914",
        num_seats=203,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:pa", "lower", 203
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/9443a886-8032-4502-b99d-551735ec9606",
        num_seats=50,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:pa", "upper", 50
        ),
    ),
)



from collections import defaultdict
from openstates.scrape import State, Organization, JurisdictionScraper
from openstates.exceptions import ScrapeValueError
import pytest


class NewJersey(State):
    pass


def test_basics():
    # id property and string
    j = NewJersey()
    assert j.jurisdiction_id == "ocd-jurisdiction/country:us/state:nj/government"
    assert j.name in str(j)


def test_as_dict():
    j = NewJersey()
    d = j.as_dict()

    assert d["_id"] == j.jurisdiction_id
    assert d["name"] == j.name
    assert d["url"] == j.url
    assert d["legislative_sessions"] == []


def test_jurisdiction_unicam_scrape():
    class Nebraska(State):
        pass

    j = Nebraska()
    js = JurisdictionScraper(j, "/tmp/")
    objects = list(js.scrape())

    # two objects, first is the Jurisdiction
    assert len(objects) == 2
    assert objects[0] == j

    # ensure we made a single legislature org
    assert isinstance(objects[1], Organization)
    assert objects[1].classification == "legislature"


def test_jurisdiction_bicameral_scrape():
    j = NewJersey()
    js = JurisdictionScraper(j, "/tmp/")
    objects = list(js.scrape())
    obj_names = set()
    obj_types = defaultdict(int)

    for o in objects:
        obj_names.add(o.name)
        obj_types[type(o)] += 1

    # ensure Jurisdiction and 3 organizations were found
    assert obj_names == {"New jersey Legislature", "Assembly", "New Jersey", "Senate"}
    assert obj_types[NewJersey] == 1
    assert obj_types[Organization] == 3


def test_jurisdiction_validate_sessions():
    j = NewJersey()
    j.legislative_sessions = [
        {
            "_scraped_name": "s1",
            "classification": "primary",
            "name": "First Session",
            "start_date": "2020-01-01",
        }
    ]
    with pytest.raises(ScrapeValueError):
        j.validate()
    j.legislative_sessions[0]["end_date"] = "2020-02-01"
    j.validate()



import pytest
from openstates.scrape import VoteEvent as ScrapeVoteEvent, Bill as ScrapeBill
from openstates.importers import VoteEventImporter, BillImporter
from openstates.data.models import (
    Jurisdiction,
    Person,
    Organization,
    Division,
    VoteEvent,
    LegislativeSession,
    Bill,
)
from openstates.utils.transformers import fix_bill_id


class DumbMockImporter(object):
    """ this is a mock importer that implements a resolve_json_id that is just a pass-through """

    def resolve_json_id(self, json_id, allow_no_match=False):
        return json_id


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    j = Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")
    j.legislative_sessions.create(name="1900", identifier="1900")
    Organization.objects.create(name="House", classification="lower", jurisdiction=j)
    Organization.objects.create(name="Senate", classification="upper", jurisdiction=j)
    return j


@pytest.mark.django_db
def test_full_vote_event():
    create_jurisdiction()
    bill = ScrapeBill("HB 1", "1900", "Axe & Tack Tax Act", chamber="lower")
    vote_event = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-01",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        chamber="lower",
    )
    vote_event.set_count("yes", 20)
    vote_event.yes("John Smith")
    vote_event.no("Adam Smith")

    Person.objects.create(name="John Smith")
    Person.objects.create(name="Adam Smith")
    for person in Person.objects.all():
        person.memberships.create(
            organization=Organization.objects.get(classification="lower")
        )

    bi = BillImporter("jid")
    bi.import_data([bill.as_dict()])

    VoteEventImporter("jid", bi).import_data([vote_event.as_dict()])

    assert VoteEvent.objects.count() == 1
    ve = VoteEvent.objects.get()
    assert ve.legislative_session == LegislativeSession.objects.get()
    assert ve.motion_classification == ["passage:bill"]
    assert ve.bill == Bill.objects.get()
    count = ve.counts.get()
    assert count.option == "yes"
    assert count.value == 20
    votes = list(ve.votes.all())
    assert len(votes) == 2
    for v in ve.votes.all():
        if v.voter_name == "John Smith":
            assert v.option == "yes"
            assert v.voter == Person.objects.get(name="John Smith")
        else:
            assert v.option == "no"
            assert v.voter == Person.objects.get(name="Adam Smith")


@pytest.mark.django_db
def test_vote_event_identifier_dedupe():
    j = create_jurisdiction()
    Organization.objects.create(
        id="org-id", name="Legislature", classification="legislature", jurisdiction=j
    )

    vote_event = ScrapeVoteEvent(
        legislative_session="1900",
        start_date="2013",
        classification="anything",
        result="passed",
        motion_text="a vote on something",
        identifier="Roll Call No. 1",
    )
    bi = BillImporter("jid")

    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "insert"
    assert VoteEvent.objects.count() == 1

    # same exact vote event, no changes
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "noop"
    assert VoteEvent.objects.count() == 1

    # new info, update
    vote_event.result = "failed"
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "update"
    assert VoteEvent.objects.count() == 1

    # new bill, insert
    vote_event.identifier = "Roll Call 2"
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "insert"
    assert VoteEvent.objects.count() == 2


@pytest.mark.django_db
def test_vote_event_dedupe_key_dedupe():
    j = create_jurisdiction()
    Organization.objects.create(
        id="org-id", name="Legislature", classification="legislature", jurisdiction=j
    )

    vote_event = ScrapeVoteEvent(
        legislative_session="1900",
        start_date="2013",
        classification="anything",
        result="passed",
        motion_text="a vote on something",
        identifier="Roll Call No. 1",
    )
    vote_event.dedupe_key = "foo"

    bi = BillImporter("jid")
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "insert"
    assert VoteEvent.objects.count() == 1

    # same exact vote event, no changes
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "noop"
    assert VoteEvent.objects.count() == 1

    # new info, update
    vote_event.result = "failed"
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "update"
    assert VoteEvent.objects.count() == 1

    # new bill identifier, update
    vote_event.identifier = "First Roll Call"
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "update"
    assert VoteEvent.objects.count() == 1

    # new identifier, insert
    vote_event.dedupe_key = "bar"
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "insert"
    assert VoteEvent.objects.count() == 2


@pytest.mark.django_db
def test_vote_event_bill_id_dedupe():
    create_jurisdiction()
    bill = Bill.objects.create(
        id="bill-1",
        identifier="HB 1",
        legislative_session=LegislativeSession.objects.get(),
        from_organization=Organization.objects.get(classification="lower"),
    )
    bill2 = Bill.objects.create(
        id="bill-2",
        identifier="HB 2",
        legislative_session=LegislativeSession.objects.get(),
        from_organization=Organization.objects.get(classification="lower"),
    )

    vote_event = ScrapeVoteEvent(
        legislative_session="1900",
        start_date="2013",
        classification="anything",
        result="passed",
        motion_text="a vote on something",
        bill=bill.identifier,
        bill_chamber="lower",
        chamber="lower",
    )
    bi = BillImporter("jid")

    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "insert"
    assert VoteEvent.objects.count() == 1

    # same exact vote event, no changes
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "noop"
    assert VoteEvent.objects.count() == 1

    # new info, update
    vote_event.result = "failed"
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "update"
    assert VoteEvent.objects.count() == 1

    # new vote event, insert
    vote_event = ScrapeVoteEvent(
        legislative_session="1900",
        start_date="2013",
        classification="anything",
        result="passed",
        motion_text="a vote on something",
        bill=bill2.identifier,
        bill_chamber="lower",
        chamber="lower",
    )
    _, what = VoteEventImporter("jid", bi).import_item(vote_event.as_dict())
    assert what == "insert"
    assert VoteEvent.objects.count() == 2


@pytest.mark.django_db
def test_vote_event_bill_clearing():
    # ensure that we don't wind up with vote events sitting around forever on bills as
    # changes make it look like there are multiple vote events
    create_jurisdiction()
    bill = Bill.objects.create(
        id="bill-1",
        identifier="HB 1",
        legislative_session=LegislativeSession.objects.get(),
        from_organization=Organization.objects.get(classification="lower"),
    )
    Bill.objects.create(
        id="bill-2",
        identifier="HB 2",
        legislative_session=LegislativeSession.objects.get(),
        from_organization=Organization.objects.get(classification="lower"),
    )
    bi = BillImporter("jid")

    vote_event1 = ScrapeVoteEvent(
        legislative_session="1900",
        start_date="2013",
        classification="anything",
        result="passed",
        motion_text="a vote on somthing",  # typo intentional
        bill=bill.identifier,
        bill_chamber="lower",
        chamber="lower",
    )
    vote_event2 = ScrapeVoteEvent(
        legislative_session="1900",
        start_date="2013",
        classification="anything",
        result="passed",
        motion_text="a vote on something else",
        bill=bill.identifier,
        bill_chamber="lower",
        chamber="lower",
    )

    # have to use import_data so postimport is called
    VoteEventImporter("jid", bi).import_data(
        [vote_event1.as_dict(), vote_event2.as_dict()]
    )
    assert VoteEvent.objects.count() == 2

    # a typo is fixed, we don't want 3 vote events now
    vote_event1.motion_text = "a vote on something"
    VoteEventImporter("jid", bi).import_data(
        [vote_event1.as_dict(), vote_event2.as_dict()]
    )
    assert VoteEvent.objects.count() == 2


@pytest.mark.django_db
def test_vote_event_bill_actions():
    create_jurisdiction()
    bill = ScrapeBill("HB 1", "1900", "Axe & Tack Tax Act", chamber="lower")

    # add actions, passage of upper & lower on same day, something else,
    # then passage in upper again on a different day
    bill.add_action(description="passage", date="1900-04-01", chamber="upper")
    bill.add_action(description="passage", date="1900-04-01", chamber="lower")
    bill.add_action(description="other event", date="1900-04-01", chamber="lower")
    bill.add_action(description="passage", date="1900-04-02", chamber="upper")

    # four passage votes, one per chamber, one on 04-01, and one on 04-02
    ve1 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-01",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        bill_action="passage",
        chamber="lower",
    )
    ve2 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-01",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        bill_action="passage",
        chamber="upper",
    )
    ve3 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-02",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        bill_action="passage",
        chamber="lower",
    )
    ve4 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-02",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        bill_action="passage",
        chamber="upper",
    )

    bi = BillImporter("jid")
    bi.import_data([bill.as_dict()])

    VoteEventImporter("jid", bi).import_data(
        [ve1.as_dict(), ve2.as_dict(), ve3.as_dict(), ve4.as_dict()]
    )

    bill = Bill.objects.get()
    votes = list(VoteEvent.objects.all())
    actions = list(bill.actions.all())
    assert len(actions) == 4
    assert len(votes) == 4

    votes = {
        (v.organization.classification, v.start_date): v.bill_action for v in votes
    }

    # ensure that votes are matched using action, chamber, and date
    assert votes[("upper", "1900-04-01")] == actions[0]
    assert votes[("lower", "1900-04-01")] == actions[1]
    assert votes[("upper", "1900-04-02")] == actions[3]
    assert votes[("lower", "1900-04-02")] is None


@pytest.mark.django_db
def test_vote_event_bill_actions_two_stage():
    # this test is very similar to what we're testing in test_vote_event_bill_actions w/
    # ve3 and ve4, that two bills that reference the same action won't conflict w/ the
    # OneToOneField, but in this case we do it in two stages so that the conflict is found
    # even if the votes weren't in the same scrape
    create_jurisdiction()
    bill = ScrapeBill("HB 1", "1900", "Axe & Tack Tax Act", chamber="lower")

    bill.add_action(description="passage", date="1900-04-02", chamber="lower")

    ve1 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-02",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        bill_action="passage",
        chamber="lower",
    )
    ve2 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-02",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        bill_action="passage",
        chamber="lower",
    )
    # disambiguate them
    ve1.dedupe_key = "one"
    ve2.dedupe_key = "two"

    bi = BillImporter("jid")
    bi.import_data([bill.as_dict()])

    # first imports just fine
    VoteEventImporter("jid", bi).import_data([ve1.as_dict()])
    votes = list(VoteEvent.objects.all())
    assert len(votes) == 1
    assert votes[0].bill_action is not None

    # when second is imported, ensure that action stays pinned to first just as it would
    # have if they were both in same import
    VoteEventImporter("jid", bi).import_data([ve1.as_dict(), ve2.as_dict()])
    votes = list(VoteEvent.objects.all())
    assert len(votes) == 2
    assert votes[0].bill_action is not None
    assert votes[1].bill_action is None


@pytest.mark.django_db
def test_vote_event_bill_actions_errors():
    create_jurisdiction()
    bill = ScrapeBill("HB 1", "1900", "Axe & Tack Tax Act", chamber="lower")

    # for this bill, two identical actions, so vote matching will fail
    bill.add_action(description="passage", date="1900-04-01", chamber="lower")
    bill.add_action(description="passage", date="1900-04-01", chamber="lower")
    # this action is good, but two votes will try to match it
    bill.add_action(description="passage", date="1900-04-02", chamber="lower")

    # will match two actions
    ve1 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-01",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        identifier="1",
        bill_action="passage",
        chamber="lower",
    )
    # will match no actions
    ve2 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-01",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        identifier="2",
        bill_action="committee result",
        chamber="lower",
    )
    # these two votes will both match the same action
    ve3 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-02",
        classification="passage:bill",
        result="pass",
        bill_chamber="lower",
        bill="HB 1",
        identifier="3",
        bill_action="passage",
        chamber="lower",
    )
    ve4 = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage-syz",
        start_date="1900-04-02",
        classification="passage:bill",
        result="fail",
        bill_chamber="lower",
        bill="HB 1",
        identifier="4",
        bill_action="passage",
        chamber="lower",
    )

    bi = BillImporter("jid")
    bi.import_data([bill.as_dict()])

    VoteEventImporter("jid", bi).import_data(
        [ve1.as_dict(), ve2.as_dict(), ve3.as_dict(), ve4.as_dict()]
    )

    bill = Bill.objects.get()
    votes = list(VoteEvent.objects.all().order_by("identifier"))

    # isn't matched, was ambiguous across two actions
    assert votes[0].bill_action is None
    # isn't matched, no match in actions
    assert votes[1].bill_action is None

    # these both try to match the same action, only first will succeed
    assert votes[2].bill_action is not None
    assert votes[3].bill_action is None


@pytest.mark.django_db
def test_fix_bill_id():
    create_jurisdiction()
    bill = ScrapeBill(
        "HB 1", "1900", "Test Bill ID", classification="bill", chamber="lower"
    )

    from openstates.settings import IMPORT_TRANSFORMERS

    IMPORT_TRANSFORMERS["bill"] = {
        "identifier": fix_bill_id,
    }

    bi = BillImporter("jid")
    bi.import_data([bill.as_dict()])

    ve = ScrapeVoteEvent(
        legislative_session="1900",
        motion_text="passage",
        start_date="1900-04-02",
        classification="passage:bill",
        result="fail",
        bill_chamber="lower",
        bill="HB1",
        identifier="4",
        bill_action="passage",
        chamber="lower",
    )

    VoteEventImporter("jid", bi).import_data([ve.as_dict()])

    IMPORT_TRANSFORMERS["bill"] = {}

    ve = VoteEvent.objects.get()
    ve.bill.identifier == "HB 1"



import pytest
from openstates.data.models import Jurisdiction, Division, Organization, Bill
from ..computed_fields import update_bill_fields


def create_data():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    j = Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")
    org = Organization.objects.create(
        jurisdiction=j, name="House", classification="lower"
    )
    j.legislative_sessions.create(identifier="1899", name="1899")
    session = j.legislative_sessions.create(identifier="1900", name="1900").id

    return session, org


@pytest.mark.django_db
def test_update_bill_basic():
    session, org = create_data()
    b = Bill.objects.create(
        identifier="HB1", title="One", legislative_session_id=session
    )
    # order doesn't matter compared to date
    b.actions.create(
        date="2020-04-20", description="Introduced", order=100, organization=org
    )
    b.actions.create(
        date="2020-04-21",
        description="Passed House",
        order=2,
        organization=org,
        classification=["passage"],
    )
    # two actions on final date, order=4 should win as latest action description
    b.actions.create(
        date="2020-04-22", description="Something Else", order=3, organization=org
    )
    b.actions.create(
        date="2020-04-22", description="Amended in Senate", order=4, organization=org
    )

    update_bill_fields(b)
    assert b.first_action_date == "2020-04-20"
    assert b.latest_action_date == "2020-04-22"
    assert b.latest_passage_date == "2020-04-21"
    assert b.latest_action_description == "Amended in Senate"



from ..models import State, Chamber, simple_numbered_districts

WI = State(
    name="Wisconsin",
    abbr="WI",
    capital="Madison",
    capital_tz="America/Chicago",
    fips="55",
    unicameral=False,
    legislature_name="Wisconsin State Legislature",
    legislature_organization_id="ocd-organization/1cba9936-eb76-4465-acb5-50c93cd4e8c9",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/791f3d9d-86c8-565a-b286-c996e8bcb17e",
    division_id="ocd-division/country:us/state:wi",
    jurisdiction_id="ocd-jurisdiction/country:us/state:wi/government",
    url="http://legis.wisconsin.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="Assembly",
        organization_id="ocd-organization/2d0dada5-8153-44aa-b436-c92a90db0a9e",
        num_seats=99,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wi", "lower", 99
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c8a757e6-378e-468c-a526-e81ca083124b",
        num_seats=33,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wi", "upper", 33
        ),
    ),
)



import pytest  # type: ignore
import datetime
from pydantic import ValidationError
from openstates.models.common import (
    validate_fuzzy_date,
    validate_ocd_jurisdiction,
    validate_ocd_person,
    validate_str_no_newline,
    validate_url,
    Link,
    OtherName,
    OtherIdentifier,
)
from openstates.models.people import (
    Person,
    Party,
    RoleType,
    OfficeType,
    Office,
    PersonIdBlock,
    Role,
)
from openstates.models.committees import (
    Membership,
    ScrapeCommittee,
    Committee,
)

VALID_PERSON_ID = "ocd-person/abcdef98-0123-7777-8888-1234567890ab"
VALID_ORG_ID = "ocd-organization/abcdef98-0123-7777-8888-1234567890ab"
VALID_JURISDICTION_ID = "ocd-jurisdiction/country:us/state:nc/government"


@pytest.mark.parametrize(
    "validator,val,valid",
    [
        (validate_fuzzy_date, "2020", True),
        (validate_fuzzy_date, "2020-01", True),
        (validate_fuzzy_date, "2020-01-22", True),
        (validate_fuzzy_date, datetime.date(2020, 1, 22), True),
        (validate_fuzzy_date, "2020-1-22", False),
        (validate_fuzzy_date, "2020/1/22", False),
        (validate_fuzzy_date, "x", False),
        (validate_ocd_person, VALID_PERSON_ID, True),
        (validate_ocd_person, "abcdef98-0123-7777-8888-1234567890ab", False),
        (validate_ocd_person, "ocd-person/abcdef980123777788881234567890ab", False),
        (
            validate_ocd_jurisdiction,
            "ocd-jurisdiction/country:us/state:nc/government",
            True,
        ),
        (validate_ocd_jurisdiction, "ocd-jurisdiction/country:us/state:nc", False),
        (
            validate_ocd_jurisdiction,
            "ocd-jurisdiction/country:us/state:xy/government",
            False,
        ),
        (
            validate_ocd_jurisdiction,
            "ocd-jurisdiction/country:us/state:nc/county:wake",
            False,
        ),
        (validate_str_no_newline, "long string with no breaks", True),
        (validate_str_no_newline, "multi\nline", False),
        (validate_url, "http://example.com", True),
        (validate_url, "https://example.com", True),
        (validate_url, "example.com", False),
    ],
)
def test_common_validators(validator, val, valid):
    if valid:
        assert validator(val) == val
    else:
        with pytest.raises(ValueError):
            validator(val)


def test_link():
    good = Link(url="https://example.com", note="simple note")
    assert good.url and good.note
    with pytest.raises(ValidationError):
        Link(url="bad-url")
    with pytest.raises(ValidationError):
        Link(url="https://good.url", note="no \n newlines!")
    with pytest.raises(ValidationError):
        Link(note="missing URL!")


def test_other_name():
    good = OtherName(name="fine", start_date="2021")
    assert good.name
    with pytest.raises(ValidationError):
        OtherName(name="newline \n not allowed!")
    with pytest.raises(ValidationError):
        OtherName(name="bad date", start_date="2")
    with pytest.raises(ValidationError):
        OtherName(name="bad date", end_date="2")
    with pytest.raises(ValidationError):
        OtherName(start_date="2021")


def test_other_ids():
    good = OtherIdentifier(identifier="fine", scheme="openstates", start_date="2021")
    assert good.identifier
    with pytest.raises(ValidationError):
        OtherIdentifier(identifier="newline \n not allowed!", scheme="openstates")
    with pytest.raises(ValidationError):
        OtherIdentifier(identifier="no scheme")
    with pytest.raises(ValidationError):
        OtherIdentifier(identifier="bad date", scheme="openstates", start_date="x")
    with pytest.raises(ValidationError):
        OtherIdentifier(identifier="bad date", scheme="openstates", end_date="x")


def test_person_basics():
    with pytest.raises(ValidationError):
        Person(name="missing fields")
    good = Person(
        id="ocd-person/11111111-2222-3333-4444-555555555555",
        name="Joan Jones",
        party=[Party(name="Democratic")],
        roles=[],
    )
    assert good.name
    with pytest.raises(ValidationError):
        good.death_date = "X"
    with pytest.raises(ValidationError):
        good.birth_date = "X"
    with pytest.raises(ValidationError):
        good.birth_date = "X"
    with pytest.raises(ValidationError):
        good.id = "123"
    with pytest.raises(ValidationError):
        good.image = "/fragment"


def test_person_commas():
    with pytest.raises(ValidationError):
        Person(
            id="ocd-person/11111111-2222-3333-4444-555555555555",
            name="Jones, Joan",
            party=[Party(name="Democratic")],
            roles=[],
        )
    good_comma = Person(
        id="ocd-person/11111111-2222-3333-4444-555555555555",
        name="Joan Jones, Jr.",
        party=[Party(name="Democratic")],
        roles=[],
    )
    assert good_comma.name


def test_party_cls():
    party = Party(name="Democratic")
    assert party.name
    with pytest.raises(ValidationError):
        party.end_date = "x"


def test_office():
    # need at least one type
    with pytest.raises(ValidationError):
        Office(classification=OfficeType.DISTRICT)
    cd = Office(classification=OfficeType.DISTRICT, address="123 Boogie Woogie Ave")

    # no newline
    with pytest.raises(ValidationError):
        cd.address = "123 Boogie Woogie Avenue\nSpringfield, MA"

    # phone number regex
    with pytest.raises(ValidationError):
        cd.voice = "911"
    with pytest.raises(ValidationError):
        cd.fax = "911"
    cd.fax = "919-555-1234"
    cd.voice = "1-123-555-6666 ext. 3333"

    # no such field
    with pytest.raises(ValueError):
        cd.phone = "911"


def test_person_id_block():
    assert PersonIdBlock(twitter="realFoolish")
    with pytest.raises(ValidationError):
        PersonIdBlock(twitter="@realFoolish")
    with pytest.raises(ValidationError):
        PersonIdBlock(youtube="https://youtube.com/test")


def test_role_basics():
    with pytest.raises(ValidationError):
        Role(type=RoleType.UPPER, jurisdiction="us")
    with pytest.raises(ValidationError):
        Role(
            type=RoleType.UPPER,
            jurisdiction=VALID_JURISDICTION_ID,
            end_reason="stuff\nhere",
        )


def test_role_conditional_requires():
    assert Role(
        type=RoleType.UPPER,
        district=4,
        end_date="2010",
        jurisdiction=VALID_JURISDICTION_ID,
    )
    assert Role(
        type=RoleType.GOVERNOR,
        start_date="2010",
        end_date="2016",
        jurisdiction=VALID_JURISDICTION_ID,
    )

    with pytest.raises(ValidationError):
        assert Role(
            type=RoleType.UPPER, end_date="2010", jurisdiction=VALID_JURISDICTION_ID
        )

    with pytest.raises(ValidationError):
        assert Role(
            type=RoleType.GOVERNOR,
            start_date="2010",
            jurisdiction=VALID_JURISDICTION_ID,
        )


def test_party_on_person():
    p = Person(
        id=VALID_PERSON_ID,
        name="Tony Tigre",
        party=[Party(name="Democratic")],
        roles=[],
    )
    with pytest.raises(ValidationError):
        # no such party
        p.party = [Party(name="Vampire")]


def test_party_required_on_legislator():
    p = Person(
        id=VALID_PERSON_ID,
        name="Tony Tigre",
        party=[Party(name="Democratic")],
        roles=[
            Role(type=RoleType.UPPER, district=1, jurisdiction=VALID_JURISDICTION_ID)
        ],
    )
    with pytest.raises(ValidationError):
        # no party!
        p.party = []


def test_multiple_parties():
    p = Person(
        id=VALID_PERSON_ID,
        name="Tony Tigre",
        party=[Party(name="Democratic")],
        roles=[],
    )
    with pytest.raises(ValidationError):
        # can't have two active major parties
        p.party = [Party(name="Democratic"), Party(name="Republican")]
    # can be in multiple parties as long as one is non-major
    p.party = [Party(name="Democratic"), Party(name="Green")]
    # or if one is obsolete
    p.party = [Party(name="Democratic", end_date="2010"), Party(name="Republican")]


def test_committee_membership():
    assert Membership(name="Franz Ferdinand", role="member")
    assert Membership(name="Franz Ferdinand", role="member", person_id=VALID_PERSON_ID)
    with pytest.raises(ValidationError):
        Membership(name="No Role", person_id=VALID_PERSON_ID)
    with pytest.raises(ValidationError):
        Membership(name="Bad ID", role="chair", person_id="123")


def test_scrapecommittee():
    assert ScrapeCommittee(name="Health", chamber="upper")
    with pytest.raises(ValidationError):
        ScrapeCommittee(name="Health \n Roads", chamber="upper")


def test_committee():
    assert Committee(
        name="Health",
        chamber="upper",
        id=VALID_ORG_ID,
        jurisdiction=VALID_JURISDICTION_ID,
    )
    with pytest.raises(ValidationError):
        Committee(
            name="Health", chamber="upper", id="123", jurisdiction=VALID_JURISDICTION_ID
        )
    with pytest.raises(ValidationError):
        Committee(
            name="Health", chamber="upper", id=VALID_ORG_ID, jurisdiction="canada"
        )


def test_committee_dict_order():
    c = Committee(
        name="Health",
        chamber="upper",
        id=VALID_ORG_ID,
        jurisdiction=VALID_JURISDICTION_ID,
    )
    assert list(c.to_dict().keys())[:4] == [
        "id",
        "jurisdiction",
        "classification",
        "name",
    ]



#!/usr/bin/env python
import typing
import os
import yaml
from pathlib import Path
from datetime import datetime
from ... import metadata
from ...models.people import Person


def add_vacancy(person: Person, until: datetime) -> None:
    with open("settings.yml") as f:
        settings = yaml.safe_load(f)
    last_role = person.roles[-1]
    abbr = metadata.lookup(jurisdiction_id=last_role.jurisdiction).abbr.lower()
    if abbr not in settings:
        settings[abbr] = {"vacancies": []}
    settings[abbr]["vacancies"].append(
        {
            "chamber": last_role.type,
            "district": last_role.district,
            "vacant_until": until.date(),
        }
    )
    with open("settings.yml", "w") as f:
        yaml.safe_dump(settings, f)


def retire_person(
    person: Person,
    end_date: str,
    reason: typing.Optional[str] = None,
    death: bool = False,
) -> tuple[Person, int]:
    num = 0
    for role in person.roles:
        if role.is_active():
            role.end_date = end_date
            if reason:
                role.end_reason = reason
            num += 1

    if death:
        person.death_date = end_date

    # remove old offices
    person.offices = []

    return person, num


def retire_file(filename: typing.Union[Path, str]) -> str:  # pragma: no cover
    if isinstance(filename, Path):
        filename = str(filename)
    new_filename = filename.replace("/legislature/", "/retired/").replace(
        "/municipalities/", "/retired/"
    )
    os.renames(filename, new_filename)
    return new_filename



import pytest
from openstates.scrape.schemas.bill import schema
from openstates.scrape.base import (
    BaseModel,
    SourceMixin,
    AssociatedLinkMixin,
)


class GenericModel(
    BaseModel,
    SourceMixin,
    AssociatedLinkMixin,
):
    """ a generic model used for testing the base and mixins """

    _type = "generic"
    _schema = schema

    def __init__(self):
        super(GenericModel, self).__init__()
        self._associated = []


def test_init_id():
    m = GenericModel()
    assert len(m._id) == 36


def test_as_dict():
    m = GenericModel()
    assert m.as_dict()["_id"] == m._id


def test_setattr():
    m = GenericModel()

    with pytest.raises(ValueError):
        m.some_random_key = 3

    # and no error raised since this is a valid key
    m._id = "new id"


def test_add_source():
    m = GenericModel()
    m.add_source("http://example.com/1")
    m.add_source("http://example.com/2", note="xyz")
    assert m.sources == [
        {"url": "http://example.com/1", "note": ""},
        {"url": "http://example.com/2", "note": "xyz"},
    ]


def test_add_associated_link_match():
    m = GenericModel()
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com/1.txt",
        media_type="text/plain",
        on_duplicate="error",
    )
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com/1.pdf",
        media_type="application/pdf",
        on_duplicate="error",
    )
    # one 'document' added, multiple links for it
    assert len(m._associated) == 1
    assert len(m._associated[0]["links"]) == 2


def test_add_associated_link_on_duplicate_bad():
    m = GenericModel()

    with pytest.raises(ValueError):
        m._add_associated_link(
            "_associated",
            "something",
            "http://example.com",
            media_type="text/html",
            on_duplicate="idk",
        )


def test_add_associated_link_on_duplicate_error():
    m = GenericModel()
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com",
        media_type="text/html",
        on_duplicate="error",
    )

    with pytest.raises(ValueError):
        m._add_associated_link(
            "_associated",
            "something else",
            "http://example.com",
            media_type="text/html",
            on_duplicate="error",
        )


def test_add_associated_link_on_duplicate_ignore():
    m = GenericModel()
    m._add_associated_link(
        "_associated",
        "something",
        "http://example.com",
        media_type="text/html",
        on_duplicate="ignore",
    )
    m._add_associated_link(
        "_associated",
        "something else",
        "http://example.com",
        media_type="text/html",
        on_duplicate="ignore",
    )
    # one 'document' added, single link for it, keeps first name
    assert len(m._associated) == 1
    assert len(m._associated[0]["links"]) == 1
    assert m._associated[0]["note"] == "something"



import re
import typing
import datetime
from enum import Enum
from pydantic import validator, root_validator
from .common import (
    BaseModel,
    TimeScoped,
    Link,
    OtherName,
    OtherIdentifier,
    validate_ocd_person,
    validate_url,
    validate_fuzzy_date,
    validate_ocd_jurisdiction,
    validate_str_no_newline,
)

# can only have one of these at a time
MAJOR_PARTIES = ("Democratic", "Republican", "Independent")

SUFFIX_RE = re.compile(r"(iii?)|(i?v)|((ed|ph|m|o)\.?d\.?)|([sj]r\.?)|(esq\.?)", re.I)
PHONE_RE = re.compile(r"^(1-)?\d{3}-\d{3}-\d{4}( ext. \d+)?$")


def validate_phone(val: str) -> str:
    if val and not PHONE_RE.match(val):
        raise ValueError("invalid phone number")
    return val


class PartyName(str, Enum):
    DEM = "Democratic"
    GREEN = "Green"
    IND = "Independent"
    LIB = "Libertarian"
    NP = "Nonpartisan"
    PROG = "Progressive"
    REP = "Republican"
    MN_DFL = "Democratic-Farmer-Labor"
    VT_DEM_PROG = "Democratic/Progressive"
    VT_PROG_DEM = "Progressive/Democratic"
    VT_REP_DEM = "Republican/Democratic"
    TN_CC_REP = "Carter County Republican"
    NY_INDEPENDENCE = "Independence"
    PR_PIP = "Partido Independentista Puertorrique\xF1o"
    PR_PNP = "Partido Nuevo Progresista"
    PR_PPD = "Partido Popular Democr\xE1tico"
    PR_PD = "Proyecto Dignidad"
    PR_MVC = "Movimiento Victoria Ciudadana"


class RoleType(str, Enum):
    UPPER = "upper"
    LOWER = "lower"
    JOINT = "legislature"
    GOVERNOR = "governor"
    LT_GOVERNOR = "lt_governor"
    MAYOR = "mayor"
    SOS = "secretary of state"
    CHIEF_ELECTION_OFFICER = "chief election officer"


LEGISLATIVE_ROLES = (
    RoleType.UPPER,
    RoleType.LOWER,
    RoleType.JOINT,
)
EXECUTIVE_ROLES = (
    RoleType.GOVERNOR,
    RoleType.LT_GOVERNOR,
    RoleType.MAYOR,
    RoleType.CHIEF_ELECTION_OFFICER,
    RoleType.SOS,
)


class OfficeType(str, Enum):
    DISTRICT = "district"
    CAPITOL = "capitol"
    PRIMARY = "primary"


class PersonIdBlock(BaseModel):
    twitter: str = ""
    youtube: str = ""
    instagram: str = ""
    facebook: str = ""

    @validator("*")
    def validate_social(cls, val: str) -> str:
        validate_str_no_newline(val)
        if val.startswith(("http://", "https://", "@")):
            raise ValueError("invalid social media account name, drop URL or @")
        return val


class Party(TimeScoped):
    name: PartyName

    _validate_strs = validator("name", allow_reuse=True)(validate_str_no_newline)


class Role(TimeScoped):
    type: RoleType
    jurisdiction: str
    district: typing.Optional[str] = None
    end_reason: str = ""  # note: this field not imported to db

    _validate_strs = validator("district", "end_reason", allow_reuse=True)(
        validate_str_no_newline
    )
    _validate_jurisdiction = validator("jurisdiction", allow_reuse=True)(
        validate_ocd_jurisdiction
    )

    @root_validator
    def check_conditional_required_fields(
        cls, values: dict[str, typing.Any]
    ) -> dict[str, typing.Any]:
        # executives require end_date, everyone else requires a district & party
        office_type = values.get("type")
        end_date = values.get("end_date")
        district = values.get("district")
        if office_type in LEGISLATIVE_ROLES and not district:
            raise ValueError("district is required for legislative roles")
        if office_type in EXECUTIVE_ROLES and not end_date:
            raise ValueError("end_date is required for executive roles")
        return values


class ScrapeOffice(BaseModel):
    classification: OfficeType
    address: str = ""
    voice: str = ""
    fax: str = ""
    name: str = ""


class Office(ScrapeOffice):
    _validate_strs = validator("address", allow_reuse=True)(validate_str_no_newline)
    _validate_phones = validator("voice", "fax", allow_reuse=True)(validate_phone)

    @property
    def display_name(self) -> str:
        return self.name or f"{self.classification.title()} Office"

    @root_validator
    def check_have_at_least_one_value(
        cls, values: dict[str, typing.Any]
    ) -> dict[str, typing.Any]:
        if not any((values.get("address"), values.get("voice"), values.get("fax"))):
            raise ValueError("must have at least one valid contact type for the office")
        return values


PARTY_ABBREVS = {
    "d": "Democratic",
    "r": "Republican",
    "dem": "Democratic",
    "rep": "Republican",
    "ind": "Independent",
    "democrat": "Democratic",
}


class ScrapePerson(BaseModel):
    name: str
    state: str
    party: PartyName
    district: str
    chamber: str
    image: str = ""
    email: str = ""
    given_name: str = ""
    family_name: str = ""
    suffix: str = ""

    links: list[Link] = []
    sources: list[Link] = []
    ids: PersonIdBlock = PersonIdBlock()
    capitol_office = ScrapeOffice(classification="capitol")
    district_office = ScrapeOffice(classification="district")
    additional_offices: list[ScrapeOffice] = []
    extras: dict = {}

    @validator("party", pre=True)
    def common_abbreviations(cls, val: str) -> str:
        # replace with proper name if one exists
        return PARTY_ABBREVS.get(val.lower(), val)

    @validator("name")
    def collapse_spaces(cls, val: str) -> str:
        return re.sub(r"\s+", " ", val).strip()

    def add_link(self, url: str, note: str = "") -> None:
        self.links.append(Link(url=url, note=note))

    def add_source(self, url: str, note: str = "") -> None:
        self.sources.append(Link(url=url, note=note))

    def add_office(
        self,
        classification: OfficeType,
        *,
        address: str = "",
        voice: str = "",
        fax: str = "",
        name: str = "",
    ) -> None:
        self.additional_offices.append(
            ScrapeOffice(
                classification=classification,
                address=address,
                voice=voice,
                fax=fax,
                name=name,
            )
        )


class Person(BaseModel):
    id: str
    name: str
    given_name: str = ""
    family_name: str = ""
    middle_name: str = ""
    suffix: str = ""
    gender: str = ""
    email: str = ""
    biography: str = ""
    birth_date: typing.Union[datetime.date, str] = ""
    death_date: typing.Union[datetime.date, str] = ""
    image: str = ""

    party: list[Party] = []
    roles: list[Role]

    offices: list[Office] = []
    links: list[Link] = []
    other_names: list[OtherName] = []
    ids: PersonIdBlock = PersonIdBlock()
    other_identifiers: list[OtherIdentifier] = []
    sources: list[Link] = []
    extras: dict = {}

    @root_validator
    def check_active_party(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        require_party = False
        for role in values.get("roles", []):
            if role.is_active() and role.type in LEGISLATIVE_ROLES:
                require_party = True

        active_parties = []
        for party in values.get("party", []):
            if party.is_active():
                active_parties.append(party.name)

        if len(active_parties) == 0 and require_party:
            raise ValueError("no active parties")
        elif len(active_parties) > 1:
            if len([party for party in active_parties if party in MAJOR_PARTIES]) > 1:
                raise ValueError(f"multiple active party memberships: {active_parties}")
            # TODO: warn again
            # else:
            #     self.warnings[person.print_filename].append(
            #         f"multiple active party memberships {active_parties}"
            #     )

        return values

    @validator("name")
    def no_bad_comma(cls, val: str) -> str:  # type: ignore
        pieces = val.split(",")
        if len(pieces) > 2:
            raise ValueError("too many commas, check if name is mangled")
        elif len(pieces) == 2 and not SUFFIX_RE.findall(pieces[1]):
            raise ValueError("invalid comma")
        return val

    _validate_dates = validator("birth_date", "death_date", allow_reuse=True)(
        validate_fuzzy_date
    )
    _validate_strings_no_newline = validator(
        # only biography is allowed newlines
        "name",
        "given_name",
        "family_name",
        "given_name",
        "middle_name",
        "email",
        "suffix",
        "gender",
        allow_reuse=True,
    )(validate_str_no_newline)
    _validate_image = validator("image", allow_reuse=True)(validate_url)

    def to_dict(self) -> dict[str, typing.Any]:
        # hack to always have id on top
        return {
            "id": self.id,
            **super().dict(exclude_defaults=True),
        }

    _validate_person_id = validator("id", allow_reuse=True)(validate_ocd_person)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openstates',
 'openstates.cli',
 'openstates.cli.tests',
 'openstates.data',
 'openstates.data.admin',
 'openstates.data.migrations',
 'openstates.data.models',
 'openstates.data.tests',
 'openstates.fulltext',
 'openstates.importers',
 'openstates.importers.tests',
 'openstates.metadata',
 'openstates.metadata._creation',
 'openstates.metadata.data',
 'openstates.metadata.tests',
 'openstates.models',
 'openstates.models.tests',
 'openstates.scrape',
 'openstates.scrape.schemas',
 'openstates.scrape.tests',
 'openstates.utils',
 'openstates.utils.people',
 'openstates.utils.tests']

package_data = \
{'': ['*'],
 'openstates.fulltext': ['raw/*'],
 'openstates.utils.tests': ['testdata/broken-committees/*',
                            'testdata/committees/*',
                            'testdata/data/*',
                            'testdata/people/data/pa/legislature/*',
                            'testdata/scraped-committees/*']}

install_requires = \
['Django>=3.2.0,<4.0.0',
 'PyYAML>=5.3.1,<6.0.0',
 'attrs>=20.2.0,<21.0.0',
 'boto3>=1.17.79,<2.0.0',
 'chardet>=3.0,<4.0',
 'click>=7.1.1,<8.0.0',
 'dj_database_url>=0.5.0,<0.6.0',
 'jsonschema>=3.2.0,<4.0.0',
 'psycopg2-binary>=2.8.4,<3.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pytz>=2019.3,<2020.0',
 'scrapelib>=2.0.2,<3.0.0',
 'textract>=1.6,<2.0',
 'us>=2.0.2,<3.0.0']

entry_points = \
{'console_scripts': ['os-committees = openstates.cli.committees:main',
                     'os-initdb = openstates.cli.initdb:main',
                     'os-people = openstates.cli.people:main',
                     'os-text-extract = openstates.cli.text_extract:main',
                     'os-update = openstates.cli.update:main',
                     'os-update-computed = openstates.cli.update_computed:main',
                     'os-us-to-yaml = openstates.cli.convert_us:main']}

setup_kwargs = {
    'name': 'openstates',
    'version': '6.8.2',
    'description': 'core infrastructure for the openstates project',
    'long_description': None,
    'author': 'James Turk',
    'author_email': 'james@openstates.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import us  # type: ignore
import typing
import csv
import uuid
import yaml
from collections import defaultdict


def calc_seats(data: dict) -> dict[str, dict[str, int]]:
    chamber_seats = {}
    for key in ("upper", "lower", "legislature"):
        seats = data.get(key + "_seats")
        if not seats:
            continue
        elif isinstance(seats, int):
            # one seat per district by default
            chamber_seats[key] = {str(s): 1 for s in range(1, seats + 1)}
        elif isinstance(seats, list):
            chamber_seats[key] = {str(s): 1 for s in seats}
        elif isinstance(seats, dict):
            chamber_seats[key] = seats
        else:  # pragma: no cover
            raise ValueError(seats)
    return chamber_seats


# loosely typed... not worth the hassle
def seats_to_args(seats: typing.Union[int, list, dict]) -> tuple[int, typing.Any]:
    if isinstance(seats, int):
        return seats, None
    elif isinstance(seats, list):
        return len(seats), {str(s): 1 for s in seats}
    else:
        return sum(seats.values()), seats


def slugify(name: str) -> str:
    return name.lower().replace(" ", "_")


def make_districts(
    parent_id: str,
    chamber_type: str,
    num: int,
    seats: typing.Any,
    division_ids: list[str],
) -> str:
    if not seats and not division_ids:
        return f"simple_numbered_districts('{parent_id}', '{chamber_type}', {num})"
    elif (
        parent_id.split(":")[-1] in ("az", "id", "nd", "nj", "wa")
        and chamber_type == "lower"
    ):
        return f"simple_numbered_districts('{parent_id}', '{chamber_type}', {num//2}, num_seats=2)"
    elif seats and not division_ids:
        prefix = "sldl" if chamber_type == "lower" else "sldu"
        return (
            "["
            + "\n".join(
                (
                    f'District("{seat}", "{chamber_type}", "{parent_id}/{prefix}:{slugify(seat)}", {num}),'
                    for seat, num in seats.items()
                )
            )
            + "]"
        )
    else:
        # seats and division_ids together
        return (
            "["
            + "\n".join(
                (
                    f'District("{seat}", "{chamber_type}", "{division_ids[seat]}", {num}),'
                    for seat, num in seats.items()
                )
            )
            + "]"
        )


if __name__ == "__main__":
    settings = yaml.load(open("metadata/_creation/settings.yml"))
    jurisdictions = csv.DictReader(open("metadata/_creation/jurisdictions.csv"))
    jurisdictions_by_name = {j["state"]: j for j in jurisdictions}
    org_ids: defaultdict[str, dict] = defaultdict(dict)
    for org in csv.DictReader(open("metadata/_creation/orgs.csv")):
        if org["jurisdiction_id"]:
            org_ids[org["jurisdiction_id"]][org["classification"]] = org["id"]

    for state in us.STATES + [us.states.lookup("PR")]:

        obj = settings[state.abbr.lower()]
        j = jurisdictions_by_name[state.name]

        leg_name = obj.pop("legislature_name")
        leg_org_id = org_ids[j["jurisdiction_id"]]["legislature"]
        try:
            exec_org_id = org_ids[j["jurisdiction_id"]]["executive"]
        except KeyError:
            # for the ones that didn't have an org id before, need a consistent id here
            # we'll just use the state name + "executive"
            namespace = uuid.UUID("877e004a-9993-5b24-8339-f83d10658258")
            exec_org_id = "ocd-organization/" + str(
                uuid.uuid5(namespace, state.name + "executive")
            )
        unicameral = state.abbr in ("DC", "NE")
        extra_import = ""

        if not unicameral:
            lower_name = obj.pop("lower_chamber_name", "House")
            lower_title = obj.pop("lower_title", "Representative")
            num_lower_seats, lower_seats = seats_to_args(obj.pop("lower_seats"))
            lower_div_ids = obj.pop("lower_division_ids", None)
            upper_name = obj.pop("upper_chamber_name", "Senate")
            upper_title = obj.pop("upper_title", "Senator")
            num_upper_seats, upper_seats = seats_to_args(obj.pop("upper_seats"))
            upper_div_ids = obj.pop("upper_division_ids", None)

            lower_ds = make_districts(
                j["division_id"], "lower", num_lower_seats, lower_seats, lower_div_ids
            )
            upper_ds = make_districts(
                j["division_id"], "upper", num_upper_seats, upper_seats, upper_div_ids
            )

            lower_org_id = org_ids[j["jurisdiction_id"]]["lower"]
            upper_org_id = org_ids[j["jurisdiction_id"]]["upper"]

            if "District" in lower_ds + upper_ds:
                extra_import += ", District"
            if "simple_numbered_districts" in lower_ds + upper_ds:
                extra_import += ", simple_numbered_districts"

            seats_block = f"""
                lower=Chamber(chamber_type="lower", name="{lower_name}", organization_id="{lower_org_id}",
                   num_seats={num_lower_seats}, title="{lower_title}", districts={lower_ds}),
                upper=Chamber(chamber_type="upper", name="{upper_name}", organization_id="{upper_org_id}",
                   num_seats={num_upper_seats}, title="{upper_title}", districts={upper_ds},),
        """
        else:
            num_leg_seats, leg_seats = seats_to_args(obj.pop("legislature_seats"))
            leg_title = obj.pop("legislature_title")
            div_ids = obj.pop("legislature_division_ids", None)
            districts = make_districts(
                j["division_id"], "legislature", num_leg_seats, leg_seats, div_ids
            )

            if "District" in districts:
                extra_import += ", District"
            if "simple_numbered_districts" in districts:
                extra_import += ", simple_numbered_districts"

            seats_block = f"""
                legislature=Chamber(chamber_type="unicameral", name="{leg_name}", organization_id="{leg_org_id}",
                     num_seats={num_leg_seats}, title="{leg_title}", districts={districts}),
                """

        # ensure we got it all
        if obj:
            raise Exception(obj)

        if state.abbr == "OR":
            fname = "metadata/data/ore.py"
        elif state.abbr == "IN":
            fname = "metadata/data/ind.py"
        else:
            fname = f"metadata/data/{state.abbr.lower()}.py"
        with open(fname, "w") as f:
            f.write(
                f"""from ..models import State, Chamber{extra_import}

{state.abbr} = State(
    name="{state.name}",
    abbr="{state.abbr}",
    capital="{state.capital}",
    capital_tz="{state.capital_tz}",
    fips="{state.fips}",
    unicameral={unicameral},
    legislature_name="{leg_name}",
    legislature_organization_id="{leg_org_id}",
    executive_name="Office of the Governor",
    executive_organization_id="{exec_org_id}",
    division_id="{j['division_id']}",
    jurisdiction_id="{j['jurisdiction_id']}",
    url="{j['url']}",
    {seats_block}
)"""
            )



from ..models import State, Chamber, simple_numbered_districts

AZ = State(
    name="Arizona",
    abbr="AZ",
    capital="Phoenix",
    capital_tz="America/Phoenix",
    fips="04",
    unicameral=False,
    legislature_name="Arizona State Legislature",
    legislature_organization_id="ocd-organization/920fc27e-e574-4700-8ada-221fdf683268",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/69fd6d9c-6730-42a7-ae22-31a4a111603c",
    division_id="ocd-division/country:us/state:az",
    jurisdiction_id="ocd-jurisdiction/country:us/state:az/government",
    url="http://www.azleg.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/b00585a8-0bd5-4a6b-bde1-1c96400f882f",
        num_seats=60,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:az", "lower", 30, num_seats=2
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c3d4c85b-888e-4188-b903-a251b957c162",
        num_seats=30,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:az", "upper", 30
        ),
    ),
)



# flake8: noqa
from .generic import (
    _make_pseudo_id,
    get_pseudo_id,
    makedirs,
    JSONEncoderPlus,
    convert_pdf,
    utcnow,
    format_datetime,
)
from .metadata import abbr_to_jid, jid_to_abbr
